# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
Bare example of how this module works::

    import k1lib

    class CbA(k1lib.Callback):
        def __init__(self):
            super().__init__()
            self.initialState = 3
        def startBatch(self):
            print("startBatch - CbA")
        def startPass(self):
            print("startPass - CbA")

    class CbB(k1lib.Callback):
        def startBatch(self):
            print("startBatch - CbB")
        def endLoss(self):
            print("endLoss - CbB")

    # initialization
    cbs = k1lib.Callbacks()
    cbs.add(CbA()).add(CbB())
    model = lambda xb: xb + 3
    lossF = lambda y, yb: y - yb

    # training loop
    cbs("startBatch"); xb = 6; yb = 2
    cbs("startPass"); y = model(xb); cbs("endPass")
    cbs("startLoss"); loss = lossF(y, yb); cbs("endLoss")
    cbs("endBatch")

    print(cbs.CbA) # can reference the Callback object directly

So, point is, you can define lots of :class:`Callback` classes that
defines a number of checkpoint functions, like ``startBatch``. Then,
you can create a :class:`Callbacks` object that includes Callback
objects. When you do ``cbs("checkpoint")``, this will execute
``cb.checkpoint()`` of all the Callback objects.

Pretty much everything here is built upon this. The core training loop
has nothing to do with ML stuff. In fact, it's just a bunch of
``cbs("...")`` statements. Everything meaningful about the training
loop comes from different Callback classes. Advantage of this is that you
can tack on wildly different functions, have them play nicely with each
other, and remove entire complex functionalities by commenting out a
single line."""
import k1lib, time, os, logging, numpy as np
plt = k1lib.dep("matplotlib.pyplot"); import k1lib.cli as cli
from typing import Set, List, Union, Callable, ContextManager, Iterator
from collections import OrderedDict
__all__ = ["Callback", "Callbacks", "Cbs"]
class Callback:                                                                  # Callback
    r"""Represents a callback. Define specific functions
inside to intercept certain parts of the training
loop. Can access :class:`k1lib.Learner` like this::

    self.l.xb = self.l.xb[None]

This takes x batch of learner, unsqueeze it at
the 0 position, then sets the x batch again.

Normally, you will define a subclass of this and
define specific intercept functions, but if you
want to create a throwaway callback, then do this::

    Callback().withCheckpoint("startRun", lambda: print("start running"))

You can use :attr:`~k1lib.callbacks.callbacks.Cbs` (automatically exposed) for
a list of default Callback classes, for any particular needs.

**order**

You can also use `.order` to set the order of execution of the callback.
The higher, the later it gets executed. Value suggestions:

- 7: pre-default runs, like LossLandscape
- 10: default runs, like DontTrainValid
- 13: custom mods, like ModifyBatch
- 15: pre-recording mod
- 17: recording mods, like Profiler.memory
- 20: default recordings, like Loss
- 23: post-default recordings, like ParamFinder
- 25: guards, like TimeLimit, CancelOnExplosion

Just leave as default (10) if you don't know what values to choose.

**dependsOn**

If you're going to extend this class, you can also specify dependencies
like this::

    class CbC(k1lib.Callback):
        def __init__(self):
            super().__init__()
            self.dependsOn = {"Loss", "Accuracy"}

This is so that if somewhere, ``Loss`` callback class is temporarily
suspended, then CbC will be suspended also, therefore avoiding errors.

**Suspension**

If your Callback is mainly dormant, then you can do something like this::

    class CbD(k1lib.Callback):
        def __init__(self):
            super().__init__()
            self.suspended = True
        def startBatch(self):
            # these types of methods will only execute
            # if ``self.suspended = False``
            pass
        def analyze(self):
            self.suspended = False
            # do something that sometimes call ``startBatch``
            self.suspended = True

    cbs = k1lib.Callbacks().add(CbD())
    # dormant phase:
    cbs("startBatch") # does not execute CbD.startBatch()
    # active phase
    cbs.CbB.analyze() # does execute CbD.startBatch()

So yeah, you can easily make every checkpoint active/dormant by changing
a single variable, how convenient. See over :meth:`Callbacks.suspend`
for more."""                                                                     # Callback
    def __init__(self):                                                          # Callback
        self.l = None; self.cbs = None; self.suspended = False                   # Callback
        self.name = self.__class__.__name__; self.dependsOn:Set[str] = set()     # Callback
        self.order = 10 # can be modified by subclasses. A smaller order will be executed first # Callback
    def suspend(self):                                                           # Callback
        """Checkpoint, called when the Callback is temporarily suspended. Overridable""" # Callback
        pass                                                                     # Callback
    def restore(self):                                                           # Callback
        """Checkpoint, called when the Callback is back from suspension. Overridable""" # Callback
        pass                                                                     # Callback
    def __getstate__(self): state = dict(self.__dict__); state.pop("l", None); state.pop("cbs", None); return state # Callback
    def __setstate__(self, state): self.__dict__.update(state)                   # Callback
    def __repr__(self): return f"{self._reprHead}, can...\n{self._reprCan}"      # Callback
    @property                                                                    # Callback
    def _reprHead(self): return f"Callback `{self.name}`"                        # Callback
    @property                                                                    # Callback
    def _reprCan(self): return """- cb.something: to get specific attribute "something" from learner if not available
- cb.withCheckpoint(checkpoint, f): to quickly insert an event handler
- cb.detach(): to remove itself from its parent Callbacks"""                     # Callback
    def withCheckpoint(self, checkpoint:str, f:Callable[["Callback"], None]):    # Callback
        """Quickly set a checkpoint, for simple, inline-able functions

:param checkpoint: checkpoints like "startRun"
:param f: function that takes in the Callback itself"""                          # Callback
        setattr(self, checkpoint, lambda: f(self)); return self                  # Callback
    def __call__(self, checkpoint):                                              # Callback
        if not self.suspended and hasattr(self, checkpoint):                     # Callback
            return getattr(self, checkpoint)() != None                           # Callback
    def attached(self):                                                          # Callback
        """Called when this is added to a :class:`Callback`. Overrides this to
do custom stuff when this happens."""                                            # Callback
        pass                                                                     # Callback
    def detach(self):                                                            # Callback
        """Detaches from the parent :class:`Callbacks`"""                        # Callback
        self.cbs.remove(self.name); return self                                  # Callback
Cbs = k1lib.Object()                                                             # Callback
Callback.lossCls = k1lib.Object()                                                # Callback
class Timings:                                                                   # Timings
    """List of checkpoint timings. Not intended to be instantiated by the end user.
Used within :class:`~k1lib.callbacks.callbacks.Callbacks`, accessible via
:attr:`Callbacks.timings` to record time taken to execute a single
checkpoint. This is useful for profiling stuff."""                               # Timings
    @property                                                                    # Timings
    def state(self):                                                             # Timings
        answer = dict(self.__dict__); answer.pop("getdoc", None); return answer  # Timings
    @property                                                                    # Timings
    def checkpoints(self) -> List[str]:                                          # Timings
        """List of all checkpoints encountered"""                                # Timings
        return [cp for cp in self.state if k1lib.isNumeric(self[cp])]            # Timings
    def __getattr__(self, attr):                                                 # Timings
        if attr.startswith("_"): raise AttributeError()                          # Timings
        self.__dict__[attr] = 0; return 0                                        # Timings
    def __getitem__(self, idx): return getattr(self, idx)                        # Timings
    def __setitem__(self, idx, value): setattr(self, idx, value)                 # Timings
    def plot(self):                                                              # Timings
        """Plot all checkpoints' execution times"""                              # Timings
        plt.figure(dpi=100); checkpoints = self.checkpoints                      # Timings
        timings = np.array([self[cp] for cp in checkpoints])                     # Timings
        maxTiming = timings.max()                                                # Timings
        if maxTiming >= 1:                                                       # Timings
            plt.bar(checkpoints, timings); plt.ylabel("Time (s)")                # Timings
        elif maxTiming >= 1e-3 and maxTiming < 1:                                # Timings
            plt.bar(checkpoints, timings*1e3); plt.ylabel("Time (ms)")           # Timings
        elif maxTiming >= 1e-6 and maxTiming < 1e-3:                             # Timings
            plt.bar(checkpoints, timings*1e6); plt.ylabel("Time (us)")           # Timings
        plt.xticks(rotation="vertical"); plt.show()                              # Timings
    def clear(self):                                                             # Timings
        """Clears all timing data"""                                             # Timings
        for cp in self.checkpoints: self[cp] = 0                                 # Timings
    def __repr__(self):                                                          # Timings
        cps = '\n'.join([f'- {cp}: {self[cp]}' for cp in self.checkpoints])      # Timings
        return f"""Timings object. Checkpoints:\n{cps}\n
Can...
- t.startRun: to get specific checkpoint's execution time
- t.plot(): to plot all checkpoints"""                                           # Timings
_time = time.time                                                                # Timings
class Callbacks:                                                                 # Callbacks
    def __init__(self):                                                          # Callbacks
        self._l: k1lib.Learner = None; self.cbsDict = {}                         # Callbacks
        self._timings = Timings(); self.contexts = [[]]                          # Callbacks
    @property                                                                    # Callbacks
    def timings(self) -> Timings:                                                # Callbacks
        """Returns :class:`~k1lib.callbacks.callbacks.Timings` object"""         # Callbacks
        return self._timings                                                     # Callbacks
    @property                                                                    # Callbacks
    def l(self) -> "k1lib.Learner":                                              # Callbacks
        """:class:`k1lib.Learner` object. Will be set automatically when
you set :attr:`k1lib.Learner.cbs` to this :class:`Callbacks`"""                  # Callbacks
        return self._l                                                           # Callbacks
    @l.setter                                                                    # Callbacks
    def l(self, learner):                                                        # Callbacks
        self._l = learner                                                        # Callbacks
        for cb in self.cbs: cb.l = learner                                       # Callbacks
    @property                                                                    # Callbacks
    def cbs(self) -> List[Callback]:                                             # Callbacks
        """List of :class:`Callback`"""                                          # Callbacks
        return [*self.cbsDict.values()] # convenience method for looping over stuff # Callbacks
    def _sort(self) -> "Callbacks":                                              # Callbacks
        self.cbsDict = OrderedDict(sorted(self.cbsDict.items(), key=(lambda o: o[1].order))); return self # Callbacks
    def add(self, cb:Callback, name:str=None):                                   # Callbacks
        """Adds a callback to the collection.
Example::

    cbs = k1lib.Callbacks()
    cbs.add(k1lib.Callback().withCheckpoint("startBatch", lambda self: print("here")))

If you just want to insert a simple callback with a single checkpoint, then
you can do something like::

    cbs.add(["startBatch", lambda _: print("here")])"""                          # Callbacks
        if isinstance(cb, (list, tuple)):                                        # Callbacks
            return self.add(Callback().withCheckpoint(cb[0], cb[1]))             # Callbacks
        if not isinstance(cb, Callback): raise RuntimeError("`cb` is not a callback!") # Callbacks
        if cb in self.cbs: cb.l = self.l; cb.cbs = self; return self             # Callbacks
        cb.l = self.l; cb.cbs = self; name = name or cb.name                     # Callbacks
        if name in self.cbsDict:                                                 # Callbacks
            i = 0                                                                # Callbacks
            while f"{name}{i}" in self.cbsDict: i += 1                           # Callbacks
            name = f"{name}{i}"                                                  # Callbacks
        cb.name = name; self.cbsDict[name] = cb; self._sort()                    # Callbacks
        self._appendContext_append(cb); cb("attached"); return self              # Callbacks
    def __contains__(self, e:str) -> bool:                                       # Callbacks
        """Whether a specific Callback name is in this :class:`Callback`."""     # Callbacks
        return e in self.cbsDict                                                 # Callbacks
    def remove(self, *names:List[str]):                                          # Callbacks
        """Removes a callback from the collection."""                            # Callbacks
        for name in names:                                                       # Callbacks
            if name not in self.cbsDict: return print(f"Callback `{name}` not found") # Callbacks
            cb = self.cbsDict[name]; del self.cbsDict[name]; cb("detached")      # Callbacks
        self._sort(); return self                                                # Callbacks
    def removePrefix(self, prefix:str):                                          # Callbacks
        """Removes any callback with the specified prefix"""                     # Callbacks
        for cb in self.cbs:                                                      # Callbacks
            if cb.name.startswith(prefix): self.remove(cb.name)                  # Callbacks
        return self                                                              # Callbacks
    def __call__(self, *checkpoints:List[str]) -> bool:                          # Callbacks
        """Calls a number of checkpoints one after another.
Returns True if any of the checkpoints return anything at all"""                 # Callbacks
        self._checkpointGraph_call(checkpoints)                                  # Callbacks
        answer = False                                                           # Callbacks
        for checkpoint in checkpoints:                                           # Callbacks
            beginTime = _time()                                                  # Callbacks
            answer |= any([cb(checkpoint) for cb in self.cbs])                   # Callbacks
            self._timings[checkpoint] += _time() - beginTime                     # Callbacks
        return answer                                                            # Callbacks
    def __getitem__(self, idx:Union[int, str]) -> Callback:                      # Callbacks
        """Get specific cbs.

:param idx: if :class:`str`, then get the Callback with this specific name,
    if :class:`int`, then get the Callback in that index."""                     # Callbacks
        return self.cbs[idx] if isinstance(idx, int) else self.cbsDict[idx]      # Callbacks
    def __iter__(self) -> Iterator[Callback]:                                    # Callbacks
        """Iterates through all :class:`Callback`."""                            # Callbacks
        for cb in self.cbsDict.values(): yield cb                                # Callbacks
    def __len__(self):                                                           # Callbacks
        """How many :class:`Callback` are there in total?"""                     # Callbacks
        return len(self.cbsDict)                                                 # Callbacks
    def __getattr__(self, attr):                                                 # Callbacks
        if attr == "cbsDict": raise AttributeError(attr)                         # Callbacks
        if attr in self.cbsDict: return self.cbsDict[attr]                       # Callbacks
        else: raise AttributeError(attr)                                         # Callbacks
    def __getstate__(self):                                                      # Callbacks
        state = dict(self.__dict__); state.pop("_l", None); return state         # Callbacks
    def __setstate__(self, state):                                               # Callbacks
        self.__dict__.update(state)                                              # Callbacks
        for cb in self.cbs: cb.cbs = self                                        # Callbacks
    def __dir__(self):                                                           # Callbacks
        answer = list(super().__dir__())                                         # Callbacks
        answer.extend(self.cbsDict.keys())                                       # Callbacks
        return answer                                                            # Callbacks
    def __repr__(self):                                                          # Callbacks
        return "Callbacks:\n" + '\n'.join([f"- {cbName}" for cbName in self.cbsDict if not cbName.startswith("_")]) + """\n
Use...
- cbs.add(cb[, name]): to add a callback with a name
- cbs("startRun"): to trigger a specific checkpoint, this case "startRun"
- cbs.Loss: to get a specific callback by name, this case "Loss"
- cbs[i]: to get specific callback by index
- cbs.timings: to get callback execution times
- cbs.checkpointGraph(): to graph checkpoint calling orders
- cbs.context(): context manager that will detach all Callbacks attached inside the context
- cbs.suspend("Loss", "Cuda"): context manager to temporarily prevent triggering checkpoints""" # Callbacks
    def withBasics(self):                                                        # Callbacks
        """Adds a bunch of very basic Callbacks that's needed for everything. Also
includes Callbacks that are not necessary, but don't slow things down"""         # Callbacks
        self.add(Cbs.CoreNormal()).add(Cbs.Profiler()).add(Cbs.Recorder())       # Callbacks
        self.add(Cbs.ProgressBar()).add(Cbs.Loss()).add(Cbs.Accuracy()).add(Cbs.DontTrainValid()) # Callbacks
        return self.add(Cbs.CancelOnExplosion()).add(Cbs.ParamFinder())          # Callbacks
    def withQOL(self):                                                           # Callbacks
        """Adds quality of life Callbacks."""                                    # Callbacks
        return self                                                              # Callbacks
    def withAdvanced(self):                                                      # Callbacks
        """Adds advanced Callbacks that do fancy stuff, but may slow things
down if not configured specifically."""                                          # Callbacks
        return self.add(Cbs.HookModule().withMeanRecorder().withStdRecorder()).add(Cbs.HookParam()) # Callbacks
@k1lib.patch(Callbacks)                                                          # Callbacks
def _resolveDependencies(self):                                                  # _resolveDependencies
    for cb in self.cbs:                                                          # _resolveDependencies
        cb._dependents:Set[Callback] = set()                                     # _resolveDependencies
        cb.dependsOn = set(cb.dependsOn)                                         # _resolveDependencies
    for cb in self.cbs:                                                          # _resolveDependencies
        for cb2 in self.cbs:                                                     # _resolveDependencies
            if cb2.__class__.__name__ in cb.dependsOn:                           # _resolveDependencies
                cb2._dependents.add(cb)                                          # _resolveDependencies
class SuspendContext:                                                            # SuspendContext
    def __init__(self, cbs:Callbacks, cbsNames:List[str], cbsClasses:List[str]): # SuspendContext
        self.cbs = cbs; self.cbsNames = cbsNames; self.cbsClasses = cbsClasses   # SuspendContext
        self.cbs.suspendStack = getattr(self.cbs, "suspendStack", [])            # SuspendContext
    def __enter__(self):                                                         # SuspendContext
        cbsClasses = set(self.cbsClasses); cbsNames = set(self.cbsNames)         # SuspendContext
        self._resolveDependencies()                                              # SuspendContext
        def explore(cb:Callback):                                                # SuspendContext
            for dept in cb._dependents:                                          # SuspendContext
                cbsClasses.add(dept.__class__.__name__); explore(dept)           # SuspendContext
        [explore(cb) for cb in self.cbs if cb.__class__.__name__ in cbsClasses or cb.name in cbsNames] # SuspendContext
        stackFrame = {cb:cb.suspended for cb in self.cbs if cb.__class__.__name__ in cbsClasses or cb.name in cbsNames} # SuspendContext
        for cb in stackFrame: cb.suspend(); cb.suspended = True                  # SuspendContext
        self.suspendStack.append(stackFrame)                                     # SuspendContext
    def __exit__(self, *ignored):                                                # SuspendContext
        for cb, oldValue in self.suspendStack.pop().items():                     # SuspendContext
            cb.suspended = oldValue; cb.restore()                                # SuspendContext
    def __getattr__(self, attr): return getattr(self.cbs, attr)                  # SuspendContext
@k1lib.patch(Callbacks)                                                          # SuspendContext
def suspend(self, *cbNames:List[str]) -> ContextManager:                         # suspend
    """Creates suspension context for specified Callbacks. Matches callbacks with
their name. Works like this::

    cbs = k1lib.Callbacks().add(CbA()).add(CbB()).add(CbC())
    with cbs.suspend("CbA", "CbC"):
        pass # inside here, only CbB will be active, and its checkpoints executed
    # CbA, CbB and CbC are all active

.. seealso:: :meth:`suspendClasses`"""                                           # suspend
    return SuspendContext(self, cbNames, [])                                     # suspend
@k1lib.patch(Callbacks)                                                          # suspend
def suspendClasses(self, *classNames:List[str]) -> ContextManager:               # suspendClasses
    """Like :meth:`suspend`, but matches callbacks' class names to the given list,
instead of matching names. Meaning::

    cbs.k1lib.Callbacks().add(Cbs.Loss()).add(Cbs.Loss())
    # cbs now has 2 callbacks "Loss" and "Loss0"
    with cbs.suspendClasses("Loss"):
        pass # now both of them are suspended"""                                 # suspendClasses
    return SuspendContext(self, [], classNames)                                  # suspendClasses
@k1lib.patch(Callbacks)                                                          # suspendClasses
def suspendEval(self, more:List[str]=[], less:List[str]=[]) -> ContextManager:   # suspendEval
    """Same as :meth:`suspendClasses`, but suspend some default classes typical
used for evaluation callbacks. Just convenience method really. Currently includes:

- HookModule, HookParam, ProgressBar
- ParamScheduler, Loss, Accuracy, Autosave
- ConfusionMatrix

:param more: include more classes to be suspended
:param less: exclude classes supposed to be suspended by default"""              # suspendEval
    classes = {"HookModule", "HookParam", "ProgressBar", "ParamScheduler", "Loss", "Accuracy", "Autosave", "ConfusionMatrix"} # suspendEval
    classes.update(more); classes -= set(less)                                   # suspendEval
    return self.suspendClasses(*classes)                                         # suspendEval
class AppendContext:                                                             # AppendContext
    def __init__(self, cbs:Callbacks, initCbs:List[Callback]=[]):                # AppendContext
        self.cbs = cbs; self.initCbs = initCbs                                   # AppendContext
    def __enter__(self):                                                         # AppendContext
        self.cbs.contexts.append([])                                             # AppendContext
        for cb in self.initCbs: self.cbs.add(cb)                                 # AppendContext
        return self.cbs                                                          # AppendContext
    def __exit__(self, *ignored):                                                # AppendContext
        [cb.detach() for cb in self.cbs.contexts.pop()]                          # AppendContext
@k1lib.patch(Callbacks)                                                          # AppendContext
def _appendContext_append(self, cb):                                             # _appendContext_append
    self.contexts[-1].append(cb)                                                 # _appendContext_append
@k1lib.patch(Callbacks)                                                          # _appendContext_append
def context(self, *initCbs:List[Callback]) -> ContextManager:                    # context
    """Add context.

Works like this::

    cbs = k1lib.Callbacks().add(CbA())
    # CbA is available
    with cbs.context(CbE(), CbF()):
        cbs.add(CbB())
        # CbA, CbB, CbE and CbF available
        cbs.add(CbC())
        # all 5 are available
    # only CbA is available

For maximum shortness, you can do this::

    with k1lib.Callbacks().context(CbA()) as cbs:
        # Cba is available
"""                                                                              # context
    return AppendContext(self, initCbs)                                          # context
@k1lib.patch(Callbacks)                                                          # context
def _checkpointGraph_call(self, checkpoints:List[str]):                          # _checkpointGraph_call
    if not hasattr(self, "_checkpointGraphDict"):                                # _checkpointGraph_call
        self._checkpointGraphDict = k1lib.Object().withAutoDeclare(lambda: k1lib.Object().withAutoDeclare(lambda: 0)) # _checkpointGraph_call
        self._lastCheckpoint = "<root>"                                          # _checkpointGraph_call
    for cp in checkpoints:                                                       # _checkpointGraph_call
        self._checkpointGraphDict[self._lastCheckpoint][cp] += 1                 # _checkpointGraph_call
        self._lastCheckpoint = cp                                                # _checkpointGraph_call
@k1lib.patch(Callbacks)                                                          # _checkpointGraph_call
def checkpointGraph(self, highlightCb:Union[str, Callback]=None):                # checkpointGraph
    """Graphs what checkpoints follows what checkpoints. Has to run at least once
first. Requires graphviz package though. Example::

    cbs = Callbacks()
    cbs("a", "b", "c", "d", "b")
    cbs.checkpointGraph() # returns graph object. Will display image if using notebooks

.. image:: ../images/checkpointGraph.png

:param highlightCb: if available, will highlight the checkpoints the callback
    uses. Can be name/class-name/class/self of callback."""                      # checkpointGraph
    g = k1lib.digraph(); s = set()                                               # checkpointGraph
    for cp1, cp1o in self._checkpointGraphDict.state.items():                    # checkpointGraph
        for cp2, v in cp1o.state.items():                                        # checkpointGraph
            g.edge(cp1, cp2, label=f"  {v}  "); s.add(cp2)                       # checkpointGraph
    if highlightCb != None:                                                      # checkpointGraph
        _cb = None                                                               # checkpointGraph
        if isinstance(highlightCb, Callback): _cb = highlightCb                  # checkpointGraph
        elif isinstance(highlightCb, type) and issubclass(highlightCb, Callback): # find cb that has the same class # checkpointGraph
            for cbo in self.cbs:                                                 # checkpointGraph
                if isinstance(cbo, highlightCb): _cb = cbo; break                # checkpointGraph
            if _cb is None: raise AttributeError(f"Can't find any Callback inside this Callbacks which is of type `{cb.__name__}`") # checkpointGraph
        elif isinstance(highlightCb, str):                                       # checkpointGraph
            for cbName, cbo in self.cbsDict.items():                             # checkpointGraph
                if cbName == highlightCb: _cb = cbo; break                       # checkpointGraph
                if type(cbo).name == highlightCb: _cb = cbo; break               # checkpointGraph
            if _cb is None: raise AttributeError(f"Can't find any Callback inside this Callbacks with name or class `{cb}`") # checkpointGraph
        else: raise AttributeError(f"Don't understand {cb}")                     # checkpointGraph
        print(f"Highlighting callback `{_cb.name}`, of type `{type(_cb)}`")      # checkpointGraph
        for cp in s:                                                             # checkpointGraph
            if hasattr(_cb, cp): g.node(cp, color="red")                         # checkpointGraph
    return g                                                                     # checkpointGraph



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
Lots of type hint mechanisms to be used by the `LLVM optimizer <llvm.html>`_
"""
import k1lib.cli as cli
import k1lib, itertools, copy, numbers; import numpy as np
from k1lib.cli.init import yieldT
from typing import List
from collections import defaultdict, deque
try: import torch; hasTorch = True
except: hasTorch = False; torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))

__all__ = ["tBase", "tAny", "tList", "tIter", "tSet", "tCollection", "tExpand",
           "tNpArray", "tTensor",
           "tListIterSet", "tListSet", "tListIter", "tArrayTypes",
           "inferType", "TypeHintException", "tLowest", "tCheck", "tOpt"]
settings = k1lib.settings.cli
settings.add("llvm", k1lib.Settings(), "settings related to LLVM-inspired optimizer `tOpt`. See more at module `k1lib.cli.typehint`")
settings.llvm.add("k1a", True, "utilize the supplementary C-compiled library automatically for optimizations")
class TypeHintException(Exception): pass                                         # TypeHintException
def klassName(self):                                                             # klassName
    if isinstance(self, tBase): return self.__class__.__name__                   # klassName
    try: return f"{self.__name__}"                                               # klassName
    except: return f"{self}"                                                     # klassName
def klassRepr(self): return f"{self}" if isinstance(self, tBase) else klassName(self) # klassRepr
class tBase:                                                                     # tBase
    def __init__(self, child=type(None)): self.child = child                     # tBase
    def __repr__(self):                                                          # tBase
        return f"<{klassName(self)} {klassRepr(self.child)}>"                    # tBase
    def __eq__(self, v):                                                         # tBase
        if not isinstance(v, tBase): return False                                # tBase
        if self.__class__ != v.__class__: return False                           # tBase
        if self.child != v.child: return False                                   # tBase
        return True                                                              # tBase
    def check(self, v):                                                          # tBase
        """Checks whether a specific object adhears to this type hint or not.
Returns :attr:`yieldT` if object does not adhere. If it does, then return the object.

Note that in the case that the object is actually an iterator, it will return
a new iterator containing all elements from the old iterator."""                 # tBase
        return NotImplemented                                                    # tBase
    def item(self):                                                              # tBase
        """Gets the child type of this type. Basically what's the type if
it were to go through :class:`~k1lib.cli.utils.item`. Example::

    # returns tTensor(torch.float32, 2)
    tTensor(torch.float32, 3).item()
"""                                                                              # tBase
        return self.child if self.child is not type(None) else tAny()            # tBase
    def expand(self, n) -> List["tBase"]:                                        # tBase
        """Expands the type to a list with ``n`` elements.
Example::

    # returns [int, int, int, int]
    tList(int).expand(4)
    # returns [int, float, float, str]
    tCollection(int, tExpand(float), str).expand(4)
"""                                                                              # tBase
        return [self.child if self.child is not type(None) else tAny()]*n        # tBase
    def __hash__(self):                                                          # tBase
        return hash(f"{self.__class__} {self.child}")                            # tBase
def checkF(t):                                                                   # checkF
    #print(t, isinstance(t, (tBase, cli.typehint.tBase)))                        # checkF
    if isinstance(t, (tBase, cli.typehint.tBase)): return t.check                # checkF
    else:                                                                        # checkF
        def inner(x):                                                            # checkF
            try: return x if isinstance(x, t) else yieldT                        # checkF
            except TypeError:                                                    # checkF
                if hasTorch and isinstance(x, torch.Tensor):                     # checkF
                    return x if x.dtype == t else yieldT                         # checkF
                return yieldT                                                    # checkF
            except Exception as e:                                               # checkF
                print(x, t); raise e                                             # checkF
        return inner                                                             # checkF
class tAny(tBase):                                                               # tAny
    def __init__(self): super().__init__()                                       # tAny
    def __repr__(self): return f"<{klassName(self)}>"                            # tAny
    def __eq__(self, v): return isinstance(v, tAny)                              # tAny
    def check(self, v): return v                                                 # tAny
    def item(self): return tAny()                                                # tAny
    def __hash__(self): return hash(f"tAny")                                     # tAny
class tIter(tBase):                                                              # tIter
    def check(self, v):                                                          # tIter
        l = []                                                                   # tIter
        for e in v:                                                              # tIter
            x = checkF(self.child)(e); l.append(x)                               # tIter
            if x == yieldT: return yieldT                                        # tIter
        return iter(l)                                                           # tIter
class tList(tBase):                                                              # tList
    def check(self, v):                                                          # tList
        if not isinstance(v, (list, tuple, range)): return yieldT                # tList
        if tIter(self.child).check(v) is yieldT: return yieldT                   # tList
        return v                                                                 # tList
class tSet(tBase):                                                               # tSet
    def check(self, v):                                                          # tSet
        if not isinstance(v, set): return False                                  # tSet
        if tIter(self.child).check(v) is yieldT: return yieldT                   # tSet
        return v                                                                 # tSet
tListIterSet = (tList, tIter, tSet)                                              # tSet
tListSet = (tList, tSet)                                                         # tSet
tListIter = (tList, tIter)                                                       # tSet
class tDict(tBase):                                                              # tDict
    def __init__(self, keys, values):                                            # tDict
        """Dictionary type.
Example::

    d = tDict(tIter(str), tIter(int))
    # returns {"a": 3} dict, so check passed
    d.check({"a": 3})"""                                                         # tDict
        super().__init__(); self.keys = keys; self.values = values               # tDict
    def check(self, v):                                                          # tDict
        if not isinstance(v, dict): return yieldT                                # tDict
        ks = self.keys.check(list(v.keys()))                                     # tDict
        vs = self.values.check(list(v.values()))                                 # tDict
        if ks is yieldT or vs is yieldT: return yieldT                           # tDict
        return {k: v for k, v in zip(ks, vs)}                                    # tDict
    def __eq__(self, v):                                                         # tDict
        if not isinstance(v, tDict): return False                                # tDict
        if self.keys != v.keys: return False                                     # tDict
        if self.values != v.values: return False                                 # tDict
        return True                                                              # tDict
    def __repr__(self):                                                          # tDict
        return f"<{klassName(self)} {klassRepr(self.keys)} {klassRepr(self.values)}>" # tDict
class tNpArray(tBase):                                                           # tNpArray
    def __init__(self, child=None, rank=None):                                   # tNpArray
        """Numpy array type.
Example::

    # returns np.array([2, 3])
    tNpArray(np.int64, 1).check(np.array([2, 3]))

:param child: the dtype of the array
:param rank: the rank/dimension of the array"""                                  # tNpArray
        super().__init__(child); self.rank = rank                                # tNpArray
    def check(self, v):                                                          # tNpArray
        if not isinstance(v, np.ndarray): return yieldT                          # tNpArray
        if self.rank is not None and self.rank != len(v.shape): return yieldT    # tNpArray
        return v                                                                 # tNpArray
    def __repr__(self): return f"<tNpArray {klassName(self.child)} rank={self.rank}>" # tNpArray
    def item(self): return (tNpArray(self.child, self.rank - 1) if self.rank > 1 else self.child) if self.rank is not None else tNpArray(self.child, None) # tNpArray
    def __eq__(self, v):                                                         # tNpArray
        if not isinstance(v, tNpArray): return False                             # tNpArray
        if self.child is not None and v.child is not None and self.child != v.child: return False # tNpArray
        if self.rank is None or v.rank is None: return True                      # tNpArray
        return self.rank == v.rank                                               # tNpArray
    def __hash__(self): return hash(f"{self.child} - {self.rank}")               # tNpArray
    def expand(self, n): return [self.item()]*n                                  # tNpArray
if hasTorch:                                                                     # tNpArray
    class tTensor(tBase):                                                        # tNpArray
        def __init__(self, child=None, rank=None):                               # tNpArray
            """PyTorch tensor type.
    Example::

        # returns torch.tensor([2.0, 3.0])
        tTensor(torch.float32, 1).check(torch.tensor([2.0, 3.0]))

    :param child: the dtype of the array
    :param rank: the rank/dimension of the tensor"""                             # tNpArray
            super().__init__(child); self.rank = rank                            # tNpArray
        def check(self, v):                                                      # tNpArray
            if not isinstance(v, torch.Tensor): return yieldT                    # tNpArray
            if self.rank is not None and self.rank != len(v.shape): return yieldT # tNpArray
            return v                                                             # tNpArray
        def __repr__(self): return f"<tTensor {klassName(self.child)} rank={self.rank}>" # tNpArray
        def item(self): return (tTensor(self.child, self.rank - 1) if self.rank > 1 else self.child) if self.rank is not None else tTensor(self.child, None) # tNpArray
        def __eq__(self, v):                                                     # tNpArray
            if not isinstance(v, tTensor): return False                          # tNpArray
            if self.child is not None and v.child is not None and self.child != v.child: return False # tNpArray
            if self.rank is None or v.rank is None: return True                  # tNpArray
            return self.rank == v.rank                                           # tNpArray
        def __hash__(self): return hash(f"{self.child} - {self.rank}")           # tNpArray
        def expand(self, n): return [self.item()]*n                              # tNpArray
    tArrayTypes = (tNpArray, tTensor)                                            # tNpArray
else:                                                                            # tNpArray
    class tTensor(tBase): pass                                                   # tNpArray
    tArrayTypes = (tNpArray,)                                                    # tNpArray
class tCollection(tBase):                                                        # tCollection
    def __init__(self, *children):                                               # tCollection
        """Fixed-length collection of things. Let's say you want a tuple with
5 values::

    a = [3, [2, 3], "e", 2.0, b'3']

Then, this would be represented like this::

    tCollection(int, tList(int), str, float, bytes)

This also works in conjunction with :class:`tExpand`, like this::

    a = [3, [2, 3], "e", 2.0, 3.0]
    tCollection(int, tList(int), str, tExpand(float))"""                         # tCollection
        super().__init__(None); self.children = list(children)                   # tCollection
        nExpands = sum(isinstance(e, tExpand) for e in children)                 # tCollection
        if nExpands > 1: raise TypeHintException("Can't have 2 `tExpand` in a `tCollection`") # tCollection
        self.nChildren = len(children) - nExpands # minimum number of children possible # tCollection
        self.expandIdx = -1                                                      # tCollection
        for i, e in enumerate(children):                                         # tCollection
            if isinstance(e, tExpand): self.expandIdx = i                        # tCollection
    def __repr__(self):                                                          # tCollection
        a = ' '.join(klassRepr(c) for c in self.children)                        # tCollection
        return f"<{klassName(self)} {a}>"                                        # tCollection
    def __eq__(self, v):                                                         # tCollection
        if isinstance(v, tCollection):                                           # tCollection
            if len(self.children) != len(v.children): return False               # tCollection
            for x, y in zip(self.children, v.children):                          # tCollection
                if x != y: return False                                          # tCollection
            return True                                                          # tCollection
        return False                                                             # tCollection
    def check(self, v):                                                          # tCollection
        t = type(v) if isinstance(v, (list, tuple)) else None                    # tCollection
        v = list(v); l = []                                                      # tCollection
        if self.expandIdx >= 0:                                                  # tCollection
            n = len(self.children); nv = len(v)                                  # tCollection
            nMatchExpand = nv-(n-1)                                              # tCollection
            for i in range(self.expandIdx):                                      # tCollection
                x = checkF(self.children[i])(v[i]); l.append(x)                  # tCollection
                if x is yieldT: return yieldT                                    # tCollection
            for i in range(self.expandIdx, self.expandIdx + nMatchExpand):       # tCollection
                x = checkF(self.children[self.expandIdx])(v[i]); l.append(x)     # tCollection
                if x is yieldT: return yieldT                                    # tCollection
            for i in range(self.expandIdx + nMatchExpand, nv):                   # tCollection
                x = checkF(self.children[i-nMatchExpand+1])(v[i]); l.append(x)   # tCollection
                if x is yieldT: return yieldT                                    # tCollection
        else:                                                                    # tCollection
            l = []                                                               # tCollection
            for c, e in zip(self.children, v):                                   # tCollection
                x = checkF(c)(e); l.append(x)                                    # tCollection
                if x is yieldT: return yieldT                                    # tCollection
        return t(l) if t else l                                                  # tCollection
    def reduce(self):                                                            # tCollection
        """Tries to reduce ``tCollection(int, int)`` to ``tIter(int)`` if possible""" # tCollection
        s = self.children[0]                                                     # tCollection
        for e in self.children:                                                  # tCollection
            if s != e: return self                                               # tCollection
        return tIter(s)                                                          # tCollection
    def item(self): return tLowest(*((t.child if isinstance(e, tExpand) else t) for t in self.children)) # tCollection
    def expand(self, n:int) -> List[tBase]:                                      # tCollection
        """Expands out this collection so that it has a specified length"""      # tCollection
        if self.expandIdx >= 0:                                                  # tCollection
            ts = []                                                              # tCollection
            for t in self.children:                                              # tCollection
                if isinstance(t, tExpand):                                       # tCollection
                    for i in range(n - len(self.children) + 1): ts.append(t.child) # tCollection
                else: ts.append(t)                                               # tCollection
            return ts                                                            # tCollection
        else:                                                                    # tCollection
            if len(self.children) == n: return list(self.children)               # tCollection
            else: # doesn't make sense, so default case should return to list of lowest child # tCollection
                return [self.item()]*n                                           # tCollection
class tExpand(tBase):                                                            # tExpand
    def __init__(self, child):                                                   # tExpand
        """Supplement to :class:`tCollection`"""                                 # tExpand
        super().__init__(child)                                                  # tExpand
    def check(self, v): return checkF(self.child)(v)                             # tExpand
settings.atomic.add("typeHint", (numbers.Number, np.number, str, bool, bytes), "atomic types used for infering type of object for optimization passes") # tExpand
def inferType(o):                                                                # inferType
    """Tries to infer the type of the input.
Example::

    # returns tList(int)
    inferType(range(10))
    # returns tTensor(torch.float32, 2)
    inferType(torch.randn(2, 3))
"""                                                                              # inferType
    if isinstance(o, range): return tList(int)                                   # inferType
    if isinstance(o, settings.atomic.typeHint): return type(o)                   # inferType
    if isinstance(o, np.ndarray): return tNpArray(o.dtype, len(o.shape))         # inferType
    if hasTorch and isinstance(o, torch.Tensor): return tTensor(o.dtype, len(o.shape)) # inferType
    if isinstance(o, (list, tuple)):                                             # inferType
        arr = []; diff = False; a = None                                         # inferType
        for e in o:                                                              # inferType
            t = inferType(e); arr.append(t)                                      # inferType
            if a is None: a = t                                                  # inferType
            if a != t: diff = True                                               # inferType
        if diff:                                                                 # inferType
            if len(arr) < 100: return tCollection(*arr)                          # inferType
            else: return tList(tLowest(*arr))                                    # inferType
        else: return tList(a)                                                    # inferType
    if isinstance(o, dict): return tDict(inferType(list(o.keys())), inferType(list(o.values()))) # inferType
    return tAny()                                                                # inferType
def lowestChild(t):                                                              # lowestChild
    if isinstance(t, tCollection): return tLowest(*t.children)                   # lowestChild
    if isinstance(t, tListIterSet): return t.child                               # lowestChild
    if isinstance(t, tArrayTypes):                                               # lowestChild
        if t.rank is None or t.rank == 1: return t.child                         # lowestChild
        if t.rank is None: return t.__class__(t.child)                           # lowestChild
        else: return t.__class__(t.child, t.rank - 1)                            # lowestChild
    raise TypeHintException(f"Type {t} does not have a lowest child")            # lowestChild
intTypes = {int, np.int8, np.int16, np.int32, np.int64, torch.int8, torch.int16, torch.int32, torch.int64} # lowestChild
floatTypes = {float, np.float16, np.float32, np.float64, torch.float16, torch.float32, torch.float64, torch.bfloat16} # lowestChild
try: floatTypes.add(np.float128) # some systems don't have float128              # lowestChild
except: pass                                                                     # lowestChild
intFloatTypes = {*intTypes, *floatTypes}                                         # lowestChild
numericTypes = {*intTypes, *floatTypes, complex, numbers.Number}                 # lowestChild
def allSame(l): return all(t == l[0] for t in l)                                 # allSame
def tLowest(*ts):                                                                # tLowest
    """Grabs the lowest possible shared type of all the example types.
Example::

    # returns tIter(float)
    tLowest(tIter(float), tList(int))"""                                         # tLowest
    # sort of like array types?                                                  # tLowest
    if all(isinstance(t, tArrayTypes) for t in ts):                              # tLowest
        lC = tLowest(*(lowestChild(t) for t in ts))                              # tLowest
        if all(isinstance(t, tTensor) for t in ts) or all(isinstance(t, tNpArray) for t in ts): # tLowest
            t = ts[0]; rank = t.rank if allSame([t.rank for t in ts]) else None  # tLowest
            child = t.child if allSame([t.child for t in ts]) else None          # tLowest
            return t.__class__(child, rank)                                      # tLowest
    # sort of like list?                                                         # tLowest
    if all(isinstance(t, (tList, tIter, tSet, *tArrayTypes, tCollection)) for t in ts): # tLowest
        lC = tLowest(*(lowestChild(t) for t in ts))                              # tLowest
        if any(isinstance(t, (tIter, tCollection)) for t in ts): return tIter(lC) # tLowest
        return tList(lC)                                                         # tLowest
    # all numeric?                                                               # tLowest
    if all(t in numericTypes for t in ts):                                       # tLowest
        if all(t in intTypes for t in ts): return int                            # tLowest
        if all(t in intFloatTypes for t in ts): return float                     # tLowest
        return numbers.Number                                                    # tLowest
    return tAny()                                                                # tLowest
def _tCheck(inp, op):                                                            # _tCheck
    a = inferType(inp); out = inp | op; b = inferType(out)                       # _tCheck
    x = checkF(a)(inp); c1 = x is yieldT                                         # _tCheck
    y = checkF(b)(out); c2 = y is yieldT                                         # _tCheck
    z = checkF(op._typehint(a))(y); c3 = z is yieldT                             # _tCheck
    if c1 or c2 or c3:                                                           # _tCheck
        global tCheckData                                                        # _tCheck
        tCheckData = [a, b, c1, c2, c3, inp, out]                                # _tCheck
        raise TypeHintException(f"Type hints are wrong. Hints: inp type ({a}), out type ({b}). Checks: {c1}, {c2}, {c3}. Inp: {inp}, out: {out}") # _tCheck
    return z                                                                     # _tCheck
class tCheck(cli.BaseCli):                                                       # tCheck
    def __init__(self):                                                          # tCheck
        """Tool similar to :class:`~k1lib.cli.trace.trace` to check whether
all type hint outputs of all clis are good or not. Example::

    assert range(1, 3) | tCheck() | item() | op()*2 == 2

Mainly used in cli unit tests. Return type of statement will be :class:`tCheck`,
which might be undesirable, so you can pipe it to :data:`yieldT` like this::

    # returns tCheck object
    range(1, 3) | tCheck() | item() | op()*2
    # returns number "2"
    range(1, 3) | tCheck() | item() | op()*2 | yieldT"""                         # tCheck
        self.inp = None                                                          # tCheck
    def __ror__(self, v): self.inp = v; return self                              # tCheck
    def __or__(self, op):                                                        # tCheck
        if op is yieldT: return self.inp                                         # tCheck
        self.inp = _tCheck(self.inp, op); return self                            # tCheck
    def __eq__(self, v): return self.inp == v                                    # tCheck
class tOpt(cli.BaseCli):                                                         # tOpt
    _passes = [];     _serialPasses = []                                         # tOpt
    _passStruct = {}; _serialStruct = {}                                         # tOpt
    n = 10                                                                       # tOpt
    def __init__(self):                                                          # tOpt
        """Optimizes clis. Let's say you have something
like this::

    range(1000) | toList() | head() | deref()

For whatever reason you forgot that you've dereferenced everything
in the middle, although you're only using 10 first elements, so the
code can't be lazy anymore. You can apply optimizations to it like this::

    range(1000) | tOpt() | toList() | head() | deref()

This will effectively turn it into this::

    range(1000) | tOpt() | head() | deref()

Normally, you'd use it in this form instead::

    # returns the optimized cli
    f = "file.txt" | tOpt() | cat() | shape(0) | tOpt
    # then you can do this to pass it through as usual
    "other file.txt" | f

Checkout the `llvm optimizer tutorial <llvm.html>` for a more in-depth explanation of this

More over, this combines nicely with :class:`~k1lib.cli.trace.trace` like this::

    range(5) | tOpt() | trace() | apply(op()**2) | deref()"""                    # tOpt
        self.inp = None; self.clis = []                                          # tOpt
        self._out = yieldT                                                       # tOpt
    @staticmethod                                                                # tOpt
    def _addBasePass(p, abstractness=1):                                         # tOpt
        """Adds an optimization pass that acts upon a single cli.
Example::

    def o1(c:BaseCli, t:tBase):
        if ...:
            return aS(lambda x: x**2)
        else:
            return None
    tOpt._addBasePass(o1, 6)
"""                                                                              # tOpt
        tOpt._passes.append([p, round(max(min(abstractness, 2), 1))])            # tOpt
        passStruct = {}                                                          # tOpt
        for a1 in range(2, 0, -1):                                               # tOpt
            passStruct[a1] = []                                                  # tOpt
            for p, a2 in tOpt._passes:                                           # tOpt
                if a2 == a1: passStruct[a1].append(p)                            # tOpt
        tOpt._passStruct = passStruct                                            # tOpt
    @staticmethod                                                                # tOpt
    def addPass(p, klasses:List[cli.BaseCli]=[], abstractness=3):                # tOpt
        """Adds an optimization pass that acts upon multiple clis in series.
Example::

    # cs: list of clis, ts: list of input type hints, 1 for each cli
    def o1(cs:List[BaseCli], ts:List[tBase], metadata={}):
        return [cs[1], cs[0]] # reorder the clis
    tOpt.addPass(o1, [toList, head], 3)

Here, we're declaring an optimization pass ``o1``. You will be given a list of cli
objects, the cli's input type hints and some extra metadata. If you can optimize
it, then you should return a list of new clis, else you should return None

Also, ``abstractness`` has varying number of legal values:
- 1-5: generic optimizations
- 6-10: analysis passes. Passes must not return anything

Higher abstraction optimizations will be called first, and then lower abstraction
optimizations will be called later. So, the idea is, just like LLVM, you can do
some analysis which will compute metadata that you can use in your optimization
passes, which will return optimized clis if it can.

Within optimization passes, you can prioritize optimizations that look at the global
picture first, before breaking the code up into tiny fragments with more detailed
optimizations, at which point it's hard to look at the global picture.

:param p: the optimization pass
:param klasses: list of cli classes in series that will trigger the pass
:param abstractness: how abstract is this optimization"""                        # tOpt
        tOpt._serialPasses.append([p, tuple(klasses), round(max(min(abstractness, 15), 1))]) # tOpt
        serialStruct = {}                                                        # tOpt
        for a1 in range(15, 0, -1):                                              # tOpt
            serialStruct[a1] = defaultdict(lambda: [])                           # tOpt
            for p, klasses, a2 in tOpt._serialPasses:                            # tOpt
                if a2 == a1: serialStruct[a1][klasses].append(p)                 # tOpt
        tOpt._serialStruct = serialStruct                                        # tOpt
    @staticmethod                                                                # tOpt
    def clearPasses():                                                           # tOpt
        """Clears all passes"""                                                  # tOpt
        tOpt._passes = [];     tOpt._serialPasses = []                           # tOpt
        tOpt._passStruct = {}; tOpt._serialStruct = {}                           # tOpt
        addSerialOpt()                                                           # tOpt
    @property                                                                    # tOpt
    def out(self):                                                               # tOpt
        if self._out == yieldT:                                                  # tOpt
            if isinstance(self.inp, cli.BaseCli):                                # tOpt
                self.clis = [self.inp, *self.clis]; self.inp = None              # tOpt
            # why wrap 2 times? We want passes to select klasses=[serial]        # tOpt
            c = cli.serial(cli.serial(*self.clis)); t = inferType(self.inp)      # tOpt
            # start optimization passes here                                     # tOpt
            for i in range(tOpt.n):                                              # tOpt
                atLeastOnce = False #print("-"*50)                               # tOpt
                for passes in tOpt._passStruct.values():                         # tOpt
                    for p in passes:                                             # tOpt
                        repl = p(c, t)                                           # tOpt
                        if repl is not None: atLeastOnce = True; c = repl # optimized version # tOpt
                if not atLeastOnce: break                                        # tOpt
            assert isinstance(c, cli.serial) and len(c.clis) == 1                # tOpt
            self._optCli = c.clis[0]; self._out = self.inp | c                   # tOpt
        return self._out                                                         # tOpt
    @property                                                                    # tOpt
    def optCli(self):                                                            # tOpt
        """Grabs the optimized cli.
Example::

    # returns optimized cli
    (range(5) | tOpt() | apply(op()**2) | deref()).optCli
    # you can also do it like this:
    range(5) | tOpt() | apply(op()**2) | deref() | tOpt.optCli
    # or even shorter like this:
    range(5) | tOpt() | apply(op()**2) | deref() | tOpt
"""                                                                              # tOpt
        self.out; return self._optCli                                            # tOpt
    def __ror__(self, it): self.inp = it; return self                            # tOpt
    def __iter__(self): return iter(self.out)                                    # tOpt
    def __or__(self, o):                                                         # tOpt
        if o is yieldT: return self.out                                          # tOpt
        if o is tOpt.optCli or o is tOpt:                                        # tOpt
            return self.optCli                                                   # tOpt
        self.clis.append(o); return self                                         # tOpt
    def __repr__(self): return f"{self.out}"                                     # tOpt
    def __eq__(self, v): return self.out == v                                    # tOpt
    def __bool__(self): return self.out                                          # tOpt
class window(cli.BaseCli):                                                       # window
    def __init__(self, n, newList=False):                                        # window
        self.n = n                                                               # window
    def __ror__(self, it):                                                       # window
        n = self.n; before = []; q = deque([], n)                                # window
        it = iter(it)                                                            # window
        for e in it:                                                             # window
            q.append(e)                                                          # window
            if len(q) == n:                                                      # window
                yield before, tuple(q), it; before.append(q.popleft())           # window
def grabTypes(cs, t):                                                            # grabTypes
    ts = [t]                                                                     # grabTypes
    for c in cs: t = c._typehint(t); ts.append(t)                                # grabTypes
    return ts                                                                    # grabTypes
def grabKlasses(iKlasses): return [type(e) for e in iKlasses]                    # grabKlasses
depth = 0; debug = False                                                         # grabKlasses
def serialOpt(c, t, metadata=None):                                              # serialOpt
    """Optimizes ``c``, which is supposed to be a :class:`~init.serial`
object, with the input type hint ``t``. If it can actually optimize it,
then it will return a new :class:`~init.serial` object, else it returns
None."""                                                                         # serialOpt
    if debug: global depth; depth += 1; print(f"serial depth: {depth}")          # serialOpt
    if metadata is None: metadata = {"route": []}                                # serialOpt
    # returns None, or a new serial object                                       # serialOpt
    if not isinstance(c, cli.serial):                                            # serialOpt
        if debug: print(f"out depth, not serial: {depth}"); depth -= 1           # serialOpt
        return None                                                              # serialOpt
    metadata["route"].append("serial")                                           # serialOpt
    cs = c.clis; ts = grabTypes(cs, t)                                           # serialOpt
    if debug: print(f"serialOpt: {[c.__class__.__name__ for c in cs]}, {ts}")    # serialOpt
    for windowSize in range(1, len(cs)+1):                                       # serialOpt
        for a, e, c in [cs, ts] | cli.transpose() | window(windowSize):          # serialOpt
            iKlasses, ths = e | cli.transpose()                                  # serialOpt
            klasses = tuple(type(e) for e in iKlasses)                           # serialOpt
            #print(klasses)                                                      # serialOpt
            for e in tOpt._serialStruct.values():                                # serialOpt
                if klasses in e:                                                 # serialOpt
                    for p in e[klasses]:                                         # serialOpt
                        res = p(iKlasses, ths, metadata)                         # serialOpt
                        #print(f"serial p: {p}, res: {0}, klasses: {klasses}")   # serialOpt
                        if res is not None:                                      # serialOpt
                            a = a | cli.toList(); c = c | cli.toList()           # serialOpt
                            if debug: print(f"out depth new: {depth}"); depth -= 1 # serialOpt
                            metadata["route"].pop()                              # serialOpt
                            return cli.serial(*(a | cli.transpose() | cli.item() if len(a) > 0 else []), # serialOpt
                                *res,                                            # serialOpt
                                *(c | cli.transpose() | cli.item() if len(c) > 0 else [])) # serialOpt
    if debug: print(f"out depth none: {depth}"); depth -= 1                      # serialOpt
    metadata["route"].pop()                                                      # serialOpt
def addSerialOpt():                                                              # addSerialOpt
    tOpt._addBasePass(serialOpt, 5)                                              # addSerialOpt
    def inner(cs, ts, metadata):                                                 # addSerialOpt
        res = serialOpt(cs[0], ts[0], metadata)                                  # addSerialOpt
        return None if res is None else [res]                                    # addSerialOpt
    tOpt.addPass(inner, [cli.serial], 15)                                        # addSerialOpt
    try: cli.optimizations.basics() # cyclic include, so mainly intended for regular use after first initialization # addSerialOpt
    except: pass                                                                 # addSerialOpt
tOpt.clearPasses();                                                              # addSerialOpt



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""`k1ui <https://github.com/157239n/k1ui>`_ is another project made in Java that
aims to record and manipulate the screen, keyboard and mouse. The interface to
that project on its own is clunky, and this module is the Python interface to
ease its use.

Not quite developed yet tho, because I'm lazy."""
import k1lib, numpy as np, asyncio, time, inspect, json, threading, dill, math, base64, os, random, warnings
k1 = k1lib; cli = k1.cli; from k1lib.cli import *; knn = k1.knn; Cbs = k1.Cbs; viz = k1.viz; websockets = k1.dep("websockets")
nn = k1.dep("torch.nn"); optim = k1.dep("torch.optim")
PIL = k1.dep("PIL"); k1.dep("graphviz"); requests = k1.dep("requests"); tf = k1.dep("torchvision.transforms")
try: import torch; hasTorch = True
except: torch = k1.dep("torch"); hasTorch = False
try: import torchvision; hasTv = True
except: hasTv = False
from typing import Callable, List, Iterator, Tuple, Union, Dict; from collections import defaultdict, deque; from functools import lru_cache
mpl = k1lib.dep("matplotlib"); plt = k1lib.dep("matplotlib.pyplot")
__all__ = ["get", "WsSession", "selectArea", "record", "execute", "Recording",
           "Track", "CharTrack", "WordTrack", "ContourTrack", "ClickTrack", "WheelTrack", "StreamTrack",
           "distNet", "TrainScreen"]
k1lib.settings.add("k1ui", k1.Settings().add("server", k1.Settings().add("http", "http://localhost:9511", "normal http server").add("ws", "ws://localhost:9512", "websocket server"), "server urls"), "docs related to k1ui java library");
settings = k1lib.settings.k1ui
settings.add("draw", k1.Settings(), "drawing settings")
settings.draw.add("trackHeight", 30, "Track's height in Recording visualization")
settings.draw.add("pad", 10, "Padding between tracks");
def get(path):                                                                   # get
    """Sends a get request to the Java server.
Example::

    k1ui.get("mouse/200/300") # move mouse to (200, 300)"""                      # get
    return requests.get(f"{settings.server.http}/{path}", timeout=60*10).text    # get
def post(path, jsObj):                                                           # post
    """Sends a post request to the Java server.
Example::

    k1ui.post("mouse/200/300") # move mouse to (200, 300)"""                     # post
    return requests.post(f"{settings.server.http}/{path}", json=jsObj, timeout=60*10).text # post
portAutoInc = k1.AutoIncrement(9520)                                             # post
class WsSession:                                                                 # WsSession
    def __init__(self, eventCb:Callable[["WsSession", dict], None], mainThreadCb:Callable[["WsSession"], None]): # WsSession
        """Creates a websocket connection with the server, with some callback functions

The callback functions (most are async btw) will be passed a WebSocket object
as the first argument. You can use it to send messages like this::

    # this will send a signal to the server to close the session
    sess.ws.send(json.dumps({"type": "close"}))
    # this will send a signal to the server requesting the current screenshot. Result will be deposited into eventCb
    sess.ws.send(json.dumps({"type": "screenshot"}))
    # this will execute a single event
    sess.ws.send(json.dumps({"type": "execute", "event": {"type": "keyTyped", "javaKeyCode": 0, ...}}))

Complete, minimum example::

    events = []
    async def eventCb(sess, event): events.append(event)
    async def mainThreadCb(sess):
        sess.stream(300) # starts a stream with output screen width of 300px
        await asyncio.sleep(2)
        await sess.ws.send(json.dumps({"type": "execute", "event": {"type": "keyPressed", "javaKeyCode": 65, "timestamp": 0}}))
        await sess.ws.send(json.dumps({"type": "execute", "event": {"type": "keyReleased", "javaKeyCode": 65, "timestamp": 0}}))
        await asyncio.sleep(10); sess.close()
    await k1ui.WsSession(eventCb, mainThreadCb).run()

What this code does is that it will communicate with the server continuously for 12
seconds, capturing all events in the mean time and save them into ``events`` list.
It will start up a UDP stream to capture screenshots continuously, and after 2 seconds,
it sends 2 events to the server, trying to type the letter "A". Finally, it waits for
another 10 seconds and then terminates the connection.

This interface is quite low-level, and is the basis for all other functionalities.
Some of them include:

* :meth:`record`: recording a session
* :meth:`execute`: executes a list of events

:param eventCb: (async) will be called whenever there's a new event
:param mainThreadCb: (async) will be called after setting up everything
:param streamWidth: specifies the width of the UDP stream, in pixels"""          # WsSession
        self.ws = None; self.eventCb = eventCb; self.mainThreadCb = mainThreadCb # WsSession
        if not inspect.iscoroutinefunction(eventCb): raise Exception(f"eventCb has to be an async function") # WsSession
        if not inspect.iscoroutinefunction(mainThreadCb): raise Exception(f"mainThreadCb has to be an async function") # WsSession
        self.closed = False; self.streams = {} # width -> [width, lock, port]    # WsSession
    async def _listenLoop(self):                                                 # WsSession
        while True:                                                              # WsSession
            res = await self.ws.recv() | cli.aS(json.loads); _type = res["type"] # WsSession
            if _type == "close": break # python sends close signal to java, java then sends a close signal back, as an acknowledgement # WsSession
            if _type == "screenshot": await self.eventCb(self, {"type": "screenshot", "bytes": base64.b64decode(res["screenshot"]), "timestamp": int(time.time()*1000)}) # WsSession
            if _type == "newEvent": await self.eventCb(self, res["event"])       # WsSession
    async def _pingLoop(self):                                                   # WsSession
        while True:                                                              # WsSession
            if self.closed: break                                                # WsSession
            try: await self.ws.send({"type": "ping"} | cli.aS(json.dumps)); await asyncio.sleep(1) # WsSession
            except: break                                                        # WsSession
    async def _streamLoop(self, width, locks, port):                             # WsSession
        import cv2; streamRefresh = 100 # refreshes udp stream after this many seconds, so that it doesn't hang # WsSession
        def threadLoop(lock, port):                                              # WsSession
            with lock, k1.captureStdout(False, True):                            # WsSession
                get(f"startStream/{width}/{port}"); cap = cv2.VideoCapture(f'udp://0.0.0.0:{port}', cv2.CAP_FFMPEG); beginTime = time.time() # WsSession
                while (cap.isOpened()):                                          # WsSession
                    if self.closed: break                                        # WsSession
                    res, frame = cap.read()                                      # WsSession
                    if not res: break                                            # WsSession
                    self.loop.create_task(self.eventCb(self, {"type": "stream", "width": width, "frame": frame[:,:,::-1], "timestamp": int(time.time()*1000)})) # WsSession
                    if time.time() - beginTime > streamRefresh + 10: break # there will be a short time (5s) where there're 2 udp streams simultaneously dumps events # WsSession
                cap.release(); get(f"stopStream/{port}")                         # WsSession
        ports = [port, port + 100]; sel = 0                                      # WsSession
        while not self.closed:                                                   # WsSession
            threading.Thread(target=threadLoop, args=(locks[sel], ports[sel])).start() # WsSession
            await asyncio.sleep(streamRefresh); sel = 1-sel                      # WsSession
    def stream(self, width):                                                     # WsSession
        """Starts a stream with a particular output width. The lower the width, the higher the fps and vice versa""" # WsSession
        if width in self.streams: raise Exception(f"Can't start stream with width {width}. Just use the existing stream.") # WsSession
        port = portAutoInc()                                                     # WsSession
        self.streams[width] = [width, [threading.Lock(), threading.Lock()], port]; import cv2 # placed here so that users can see error message if cv2 is not imported # WsSession
        asyncio.create_task(self._streamLoop(*self.streams[width]))              # WsSession
    async def run(self):                                                         # WsSession
        """Connects with Java server, set things up and runs ``mainThreadCb``""" # WsSession
        async with websockets.connect(settings.server.ws, max_size=1_000_000_000) as ws: # WsSession
            self.ws = ws; self.loop = asyncio.get_event_loop()                   # WsSession
            _listenLoop = asyncio.create_task(self._listenLoop())                # WsSession
            _pingLoop = asyncio.create_task(self._pingLoop());                   # WsSession
            try: await self.mainThreadCb(self)                                   # WsSession
            except asyncio.CancelledError: self.close()                          # WsSession
            await _listenLoop                                                    # WsSession
    def close(self):                                                             # WsSession
        """Closes the connection with the Java server"""                         # WsSession
        if self.closed: print("Already closed"); return                          # WsSession
        self.closed = True; asyncio.create_task(self.ws.send({"type": "close"} | cli.aS(json.dumps))) # WsSession
        for width, locks, port in self.streams.values():                         # WsSession
            with locks[0]: # make sure all locks are freed. Also important to have the 2 locks be nested in each other, in case everything aligns just right that evades this mechanism # WsSession
                with locks[1]: pass                                              # WsSession
    async def execute(self, events):                                             # WsSession
        """Executes a series of events"""                                        # WsSession
        events = events | sortF(op()["timestamp"]) | aS(list)                    # WsSession
        deltaT = int(time.time()*1000) - events[0]["timestamp"]                  # WsSession
        for e in events | apply(lambda x: {**x, "timestamp": x["timestamp"]+deltaT}): # WsSession
            st = e["timestamp"]/1000 - time.time()                               # WsSession
            if st > 0: await asyncio.sleep(st)                                   # WsSession
            await self.ws.send(json.dumps({"type": "execute", "event": e}))      # WsSession
def selectArea(x, y, w, h):                                                      # selectArea
    """Selects an area on the screen to focus into"""                            # selectArea
    return get(f"selectArea/{x}/{y}/{w}/{h}")                                    # selectArea
async def record(t=None, keyCode=None, streamWidth=300, f=iden()):               # selectArea
    """Records activities.
Examples::

    events = await k1ui.record(t=5) # records for 5 seconds
    events = await k1ui.record(keyCode=5) # records until "Escape" is pressed
    events = await k1ui.record() # records until interrupt signal is sent to the process

Note: these examples only work on jupyter notebooks. For regular Python processes,
check out official Python docs (https://docs.python.org/3/library/asyncio-task.html)

:param t: record duration
:param keyCode: key to stop the recording
:param streamWidth: whether to opens the UDP stream and capture screenshots at this width or not
:param f: extra event post processing function"""                                # selectArea
    events = []                                                                  # selectArea
    async def eventCb(sess, event):                                              # selectArea
        res = f(event)                                                           # selectArea
        if res is not None: events.append(res)                                   # selectArea
        if event["type"] == "keyReleased" and event["keyCode"] == keyCode: sess.close() # selectArea
    async def mainThreadCb(sess):                                                # selectArea
        if streamWidth: sess.stream(streamWidth)                                 # selectArea
        if t is not None: await asyncio.sleep(t); sess.close()                   # selectArea
        else: await asyncio.sleep(1e9)                                           # selectArea
    await WsSession(eventCb, mainThreadCb).run(); return events                  # selectArea
async def execute(events:List[dict]):                                            # selectArea
    """Executes some events"""                                                   # selectArea
    async def eventCb(sess, event): pass                                         # selectArea
    async def mainThreadCb(sess): await sess.execute(events); sess.close()       # selectArea
    await WsSession(eventCb, mainThreadCb).run()                                 # selectArea
uuid = k1.AutoIncrement(random.randint(0, int(1e9)), prefix="k1ui-")             # selectArea
def escapeHtml(s): return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;") # escapeHtml
class HtmlView:                                                                  # HtmlView
    def __init__(self, html): self.html = html                                   # HtmlView
    def _repr_html_(self): return self.html                                      # HtmlView
class Recording:                                                                 # Recording
    def __init__(self, events):                                                  # Recording
        self.uuid = uuid(); self._tracks = []                                    # Recording
        if len(events) == 0: return # shortcut to initialize using cloned tracks rather than events # Recording
        events = events | sortF(op()["timestamp"]) | deref()                     # Recording
        self._tracks.extend(ContourTrack.parse(events))                          # Recording
        self._tracks.extend(CharTrack.parse(events))                             # Recording
        self._tracks.extend(ClickTrack.parse(events))                            # Recording
        self._tracks.extend(WheelTrack.parse(events))                            # Recording
        self._tracks.extend(StreamTrack.parse(events))                           # Recording
        self._tracks = self._tracks | filt(op()) | apply(lambda x: x._rec(self)) | deref() # Recording
        self._resetTimes(); self._resetDis()                                     # Recording
    def _resetTimes(self): self.startTime, self.endTime = self._tracks | op().timeUnix().all() | joinStreams() | filt(op()) | toMin() & toMax(); return self # Recording
    def _resetDis(self): self.dis1, self.dis2 = (self.startTime+self.endTime)/2 | aS(lambda x: [x-self.duration*0.53, x+self.duration*0.53]); return self # display times # Recording
    @property                                                                    # Recording
    def duration(self): return self.endTime - self.startTime                     # Recording
    def addTracks(self, *tracks) -> "Recording":                                 # Recording
        """Adds tracks to the Recording"""                                       # Recording
        if not isinstance(tracks[0], Track) and len(tracks) == 1: tracks = tracks[0] # Recording
        self._tracks.extend(tracks | apply(lambda tr: tr._rec(self))); self._resetTimes(); self._resetDis(); return self # Recording
    def removeTracks(self, *tracks) -> "Recording":                              # Recording
        """Removes tracks from the Recording"""                                  # Recording
        if not isinstance(tracks[0], Track) and len(tracks) == 1: tracks = tracks[0] # Recording
        tracks | apply(self._tracks.remove) | ignore(); self._resetTimes(); self._resetDis(); return self # Recording
    def _normTime(self, t=None, default=None): return default if t is None else t + self.startTime # Recording
    def zoom(self, t1=None, t2=None):                                            # Recording
        """Zooms into a particular time range. If either bounds are not
specified, they will default to the start and end of all events.

:param t1: time values are relative to the recording's start time"""             # Recording
        _dis1 = self.dis1; t1 = _dis1 if t1 is None else t1 + self.startTime     # Recording
        _dis2 = self.dis2; t2 = _dis2 if t2 is None else t2 + self.startTime     # Recording
        delta = t2-t1; t1-=delta*0.03; t2+=delta*0.03; self.dis1 = t1; self.dis2 = t2 # Recording
        html = self._repr_html_(); self.dis1 = _dis1; self.dis2 = _dis2; return HtmlView(html) # Recording
    def sel(self, t1=None, t2=None, klass=None) -> List["Track"]:                # Recording
        """Selects a subset of tracks using several filters.

For selecting time, assuming we have a track that looks like
this (x, y are t1, t2)::

    # |-1--|   |-2-|
    #    |---3---|
    #  x     y

Then, tracks 1 and 3 are selected. Time values are relative to
recording's start time

:param t1: choose tracks that happen after this time
:param t2: choose tracks that happen before this time
:param klass: choose specific track class"""                                     # Recording
        tracks = self._tracks                                                    # Recording
        if klass: tracks = tracks | instanceOf(klass)                            # Recording
        if t1 is not None or t2 is not None:                                     # Recording
            t1 = self._normTime(t1, self.startTime); t2 = self._normTime(t2, self.endTime) # Recording
            tracks = tracks | apply(lambda o: [o.startTime or 0, o.endTime, o]) | ~filt(op()[1]<t1) | ~filt(op()[0]>=t2) | cut(2) # Recording
        return tracks | aS(list)                                                 # Recording
    def sel1(self, **kwargs) -> List["Track"]:                                   # Recording
        """Like :meth:`sel`, but this time gets the first element only."""       # Recording
        return self.sel(**kwargs) | item()                                       # Recording
    def time0(self) -> List[float]:                                              # Recording
        """Start and end recording times. Start time is zero"""                  # Recording
        return [0, self.endTime - self.startTime]                                # Recording
    def timeUnix(self) -> List[float]:                                           # Recording
        """Start and end recording times. Both are absolute unix times"""        # Recording
        return [self.startTime, self.endTime]                                    # Recording
    def events(self) -> List[dict]:                                              # Recording
        """Reconstructs events from the Recording's internal data.
The events are lossy though::

    events = ... # events recorded
    r = k1ui.Recording(events)
    assert r.events() != events # this is the lossy part. Don't expect the produced events match exactly with each other""" # Recording
        return self._tracks | op().events().all() | joinStreams() | sortF(op()["timestamp"]) | deref(igT=False) # Recording
    def copy(self) -> "Recording":                                               # Recording
        """Creates a clone of this recording"""                                  # Recording
        return Recording([]).addTracks(self._tracks | op().copy().all())._resetDis() # Recording
    def _repr_html_(self): return self | aS(createTrackss) | aS(drawTrackss)     # Recording
class Track:                                                                     # Track
    def __init__(self, startTime, endTime):                                      # Track
        """Time values are absolute unix time."""                                # Track
        self.recording = None; self.startTime = startTime if startTime else None; self.endTime = endTime; self.uuid = uuid() # Track
    def time0(self) -> List[float]:                                              # Track
        """Start and end track times. Times are relative to track's start time""" # Track
        return [0, self.endTime - self.startTime]                                # Track
    def time0Rec(self) -> List[float]:                                           # Track
        """Start and end track times. Times are relative to recording's start time""" # Track
        return [self.startTime-self.recording.startTime if self.startTime else None, self.endTime-self.recording.startTime] # Track
    def timeUnix(self) -> List[float]:                                           # Track
        """Start and end track times. Times are absolute unix times"""           # Track
        return [self.startTime, self.endTime]                                    # Track
    def concurrent(self) -> List["Track"]:                                       # Track
        """Grabs all tracks that are concurrent to this track"""                 # Track
        return self.recording.sel(*self.time0Rec())                              # Track
    def _rec(self, recording): self.recording = recording; return self # inject dependency # Track
    def _tooltip(self, ctx): return ""                                           # Track
    def _displayTimes(self): # shortcut func for displaying in __repr__          # Track
        s = f"{self.startTime-self.recording.startTime:.2f}s" if self.startTime else None # Track
        e = f"{self.endTime-self.recording.startTime:.2f}s"; return f"time ({s}->{e})" # Track
    def events(self) -> List[dict]:                                              # Track
        """Reconstructs events from the Track's internal data, to be implemented by subclasses.""" # Track
        return NotImplemented                                                    # Track
    def copy(self):                                                              # Track
        """Creates a clone of this Track, to be implemented by subclasses"""     # Track
        return NotImplemented                                                    # Track
    def move(self, deltaTime):                                                   # Track
        """Moves the entire track left or right, to be implemented by subclasses.

:param deltaTime: if negative, move left by this number of seconds, else move right""" # Track
        self.startTime += deltaTime; self.endTime += deltaTime; self.recording._resetTimes(); self.recording._resetDis() # Track
class CharTrack(Track):                                                          # CharTrack
    def __init__(self, keyText:str, keyCode:int, mods:List[bool], times:List[float]): # CharTrack
        """Representing 1 key pressed and released.

:param keyText: text to display to user, like "Enter"
:param keyCode: event's "javaKeyCode"
:param mods: list of 3 booleans, whether ctrl, shift or alt is pressed"""        # CharTrack
        super().__init__(*times); self.keyText = keyText; self.keyCode = keyCode; self.mods = mods # CharTrack
    @staticmethod                                                                # CharTrack
    def parse(events) -> List["CharTrack"]:                                      # CharTrack
        stacks = {} # keyCode -> obj                                             # CharTrack
        def process(e):                                                          # CharTrack
            _type, keyText, keyCode, mods, timestamp = e                         # CharTrack
            if _type == "keyPressed":                                            # CharTrack
                if keyCode in stacks and stacks[keyCode]:                        # CharTrack
                    a = stacks[keyCode]; stacks[keyCode] = e                     # CharTrack
                    return [a, [_type, keyText, keyCode, mods, timestamp - 0.001]] # CharTrack
                    #raise Exception("Strange case. Why would the same key be pressed twice without being released first") # CharTrack
                stacks[keyCode] = e                                              # CharTrack
            if _type == "keyReleased":                                           # CharTrack
                a = stacks[keyCode] if keyCode in stacks and stacks[keyCode] else None # CharTrack
                stacks[keyCode] = None; return [a, e]                            # CharTrack
        def makeTrack(x, y):                                                     # CharTrack
            if x is None: x = [0, y[1], y[2], y[3], None]                        # CharTrack
            return CharTrack(x[1], x[2], x[3], [x[4], y[4]])                     # CharTrack
        return events | filt(op()["type"].startswith("key")) | filt(op()["type"] != "keyTyped") | apply(lambda x: [x["type"], x["keyText"], x["javaKeyCode"], [x["ctrl"], x["shift"], x["alt"]], x["timestamp"]/1000]) | apply(process) | filt(op()) | ~apply(makeTrack) | deref() # CharTrack
    def _tooltip(self, ctx): return escapeHtml(self.__repr__())                  # CharTrack
    def __repr__(self): return f"<CharTrack {self._displayTimes()} keyText ({self.keyText})>" # CharTrack
    def events(self):                                                            # CharTrack
        d = []; t1, t2 = self.timeUnix() # does not care about mods because the mods will have a separate CharTrack already, so we don't have to repeat # CharTrack
        if t1: d.append({"type": "keyPressed", "keyText": self.keyText, "javaKeyCode": self.keyCode, "timestamp": int(t1*1000)}) # CharTrack
        if t2: d.append({"type": "keyReleased", "keyText": self.keyText, "javaKeyCode": self.keyCode, "timestamp": int(t2*1000)}) # CharTrack
        return d                                                                 # CharTrack
    def copy(self): return CharTrack(self.keyText, self.keyCode, self.mods, self.timeUnix()) # CharTrack
    def move(self, deltaTime):                                                   # CharTrack
        if self.startTime: self.startTime += deltaTime                           # CharTrack
        self.endTime += deltaTime; self.recording._resetTimes()                  # CharTrack
def _ord2(x):                                                                    # _ord2
    y = x | apply(ord) | deref()                                                 # _ord2
    x2y = [x, y] | toDict(False)                                                 # _ord2
    y2x = [y, x] | toDict(False)                                                 # _ord2
    return [x, y, x2y, y2x]                                                      # _ord2
_upper, _upperCs, _upperD1, _upperD2 = _ord2("ABCDEFGHIJKLMNOPQRSTUVWXYZ");      # _ord2
_lower, _lowerCs, _lowerD1, _lowerD2 = _ord2("abcdefghijklmnopqrstuvwxyz")       # _ord2
_num, _numCs, _numD1, _numD2 = _ord2("1234567890")                               # _ord2
_puncLower, _puncLowerCs, _puncLowerD1, _puncLowerD2 = _ord2("[];',./`-=\\")     # _ord2
_puncUpper, _puncUpperCs, _puncUpperD1, _puncUpperD2 = _ord2("{}:\"<>?~_+|")     # _ord2
# maps from numbers 12345 to punctuation like !@#$%                              # _ord2
_numPunc, _numPuncCs, _numPuncD1, _numPuncD2 = _ord2("!@#$%^&*()")               # _ord2
_numPuncMap1 = [_numPuncCs, _numCs] | toDict(False); _numPuncMap2 = [_numCs, _numPuncCs] | toDict(False) # _ord2
_punc, _puncCs, _puncD1, _puncD2 = _ord2(_puncLower + _puncUpper + _numPunc + " ") # _ord2
# maps from lower case punctuation like ;',./ into upper case like :"<>?         # _ord2
_puncMap = [_puncLower, _puncUpper] | toDict(False); _puncMapCs = [_puncLowerCs, _puncUpperCs] | toDict(False) # _ord2
_puncMap2 = [_puncUpper, _puncLower] | toDict(False)                             # _ord2
def _inferText(code:int, mods) -> str:                                           # _inferText
    if mods[0] or mods[2]: return None                                           # _inferText
    shift = mods[1]                                                              # _inferText
    if shift:                                                                    # _inferText
        if code in _upperCs: return _upperD2[code]                               # _inferText
        if code in _lowerCs: return _lowerD2[code].upper()                       # _inferText
        if code in _numCs: return _numPuncD2[_numPuncMap2[code]]                 # _inferText
        if code in _puncLowerCs: return _puncUpperD2[_puncMapCs[code]]           # _inferText
        if code in _puncCs: return _puncD2[code]                                 # _inferText
        return None                                                              # _inferText
    else:                                                                        # _inferText
        if code in _upperCs: return _upperD2[code].lower()                       # _inferText
        if code in _lowerCs: return _lowerD2[code]                               # _inferText
        if code in _numCs: return _numD2[code]                                   # _inferText
        if code in _puncCs: return _puncD2[code]                                 # _inferText
        return None                                                              # _inferText
def _isUpper(x:str) -> bool: return x in _upper or x in _puncUpper or x in _numPunc # _isUpper
def _canon(x:str) -> Union[int, str]: # returns canonical key to be pressed      # _canon
    if x in _num: return _numD1[x]                                               # _canon
    if x in _upper: return _upperD1[x]                                           # _canon
    if x in _lower: return _upperD1[x.upper()]                                   # _canon
    if x in _puncLower: return x                                                 # _canon
    if x in _puncUpper: return _puncMap2[x]                                      # _canon
    if x in _numPunc: return _numPuncMap1[_numPuncD1[x]]                         # _canon
    if x in _punc: return x                                                      # _canon
    return None                                                                  # _canon
def _textToKeys(text:str): # opposite of _interText                              # _textToKeys
    cap = False; d = []; sk = 16 # shift key                                     # _textToKeys
    for c in text:                                                               # _textToKeys
        _cap = _isUpper(c)                                                       # _textToKeys
        if _cap and not cap:   d.append(["down", sk]); cap = True # change to upper # _textToKeys
        elif not _cap and cap: d.append(["up", sk]); cap = False # change to lower # _textToKeys
        d.append(["down", _canon(c)]); d.append(["up", _canon(c)])               # _textToKeys
    if cap: d.append(["up", sk])                                                 # _textToKeys
    return d                                                                     # _textToKeys
def _getTextBlocks(charTracks:List["CharTrack"]): # Get potential collection of CharTracks # _getTextBlocks
    es = charTracks | filt(op().startTime) | sortF(op().startTime) | apply(lambda x: [_inferText(x.keyCode, x.mods), x]) | aS(list) # _getTextBlocks
    d = []; _d = []; inBlock = False                                             # _getTextBlocks
    for c, obj in es:                                                            # _getTextBlocks
        if c is None and inBlock: d.append(_d); inBlock = False # ends a block   # _getTextBlocks
        elif c is not None and not inBlock: _d = []; inBlock = True # starts a new block # _getTextBlocks
        if inBlock: _d.append([c, obj])                                          # _getTextBlocks
    if inBlock: d.append(_d)                                                     # _getTextBlocks
    return d | apply(transpose() | join("") + iden())                            # _getTextBlocks
class WordTrack(Track):                                                          # WordTrack
    def __init__(self, text, times:List[float]):                                 # WordTrack
        """Representing normal text input. This is not created from events
directly. Rather, it's created from scanning over CharTracks and merging them together""" # WordTrack
        super().__init__(*times); self.text = text                               # WordTrack
    def _tooltip(self, ctx): return escapeHtml(self.__repr__())                  # WordTrack
    def __repr__(self): return f"<WordTrack {self._displayTimes()} text ({self.text}) >" # WordTrack
    def events(self):                                                            # WordTrack
        es = _textToKeys(self.text); d = []; ts = np.linspace(*self.timeUnix(), len(es)) # WordTrack
        for t, (_type, code) in zip(ts, es):                                     # WordTrack
            _type = "keyPressed" if _type == "down" else "keyReleased"; t = int(t*1000) # WordTrack
            if isinstance(code, str): d.append({"type": _type, "text": code, "timestamp": t}) # WordTrack
            else: d.append({"type": _type, "javaKeyCode": code, "timestamp": t}) # WordTrack
        return d                                                                 # WordTrack
    def copy(self): return WordTrack(self.text, self.timeUnix())                 # WordTrack
@k1.patch(Recording)                                                             # WordTrack
def formWords(self) -> Recording:                                                # formWords
    """Tries to merge nearby CharTracks together that looks like the user
is trying to type something, if they make sense. Assuming the user types
"a", then "b", then "c". This should be able to detect the intent that
the user is trying to type "abc", and replace 3 CharTracks with a WordTrack.
Example::

    # example recording, run in notebook cell to see interactive interface
    r = k1ui.Recording.sample(); r
    # run in another notebook cell and compare difference
    r.formWords()"""                                                             # formWords
    for word, charTracks in _getTextBlocks(self.sel(klass=CharTrack)):           # formWords
        if len(word) <= 0: continue                                              # formWords
        ts = charTracks | op().timeUnix().all() | joinStreams() | toMin() & toMax() | deref() # formWords
        self.removeTracks(charTracks); self.addTracks(WordTrack(word, ts))       # formWords
        self.removeTracks(self.sel(*ts | apply(op()-self.startTime), klass=CharTrack) | filt(op().keyCode == 16)) # removing shift CharTracks # formWords
    return self                                                                  # formWords
class ContourTrack(Track): # mouse movements                                     # ContourTrack
    def __init__(self, coords):                                                  # ContourTrack
        """Representing mouse trajectory ("mouseMoved" event).

:param coords: numpy array with shape (#events, [x, y, unix time])"""            # ContourTrack
        super().__init__(*coords | cut(2) | toMin() & toMax()); self.coords = coords; self._cachedImg = None # ContourTrack
    @staticmethod                                                                # ContourTrack
    def parse(events) -> List["ContourTrack"]:                                   # ContourTrack
        coords = events | filt(lambda x: x["type"] == "mouseMoved" or x["type"] == "mouseDragged") | apply(lambda x: [x["x"], x["y"], x["timestamp"]/1000]) | deref() | aS(np.array) # ContourTrack
        return [] if coords | shape(0) == 0 else [ContourTrack(coords)]          # ContourTrack
    def _img(self):                                                              # ContourTrack
        if self._cachedImg: return self._cachedImg                               # ContourTrack
        x, y, t = self.coords | transpose(); c = mpl.cm.rainbow(t - t[0] | aS(lambda x: x/x[-1])); plt.scatter(x, y, None, c, ".") # ContourTrack
        plt.colorbar(mpl.cm.ScalarMappable(norm=mpl.colors.Normalize(*self.time0Rec()), cmap=mpl.cm.rainbow)).ax.set_title("Time (s)") # ContourTrack
        plt.title("ContourTrack"); plt.grid(True); plt.tight_layout(); self._cachedImg = plt.gcf() | toImg(); return self._cachedImg # ContourTrack
    def __repr__(self): return f"<ContourTrack {self._displayTimes()} n ({self.coords.shape[0]})>" # ContourTrack
    def _tooltip(self, ctx):                                                     # ContourTrack
        return f"""<div><div style="margin-bottom:10px">{escapeHtml(self.__repr__())}</div>{self._imgHtml()}</div>""" # ContourTrack
    def _imgHtml(self): return f"""<img src="data:image/png;base64,{self._img() | toBytes(dataType="png") | aS(base64.b64encode) | op().decode()}" alt="Mouse trajectory" />""" # ContourTrack
    def _repr_html_(self): return f"""<!-- k1ui.ContourTrack --><div>{self._imgHtml()}</div>""" # ContourTrack
    def events(self): return self.coords | ~apply(lambda x, y, t: {"type": "mouseMoved", "x": x, "y": y, "timestamp": int(t*1000)}) | deref() # ContourTrack
    def copy(self): return ContourTrack(np.copy(self.coords))                    # ContourTrack
    def move(self, deltaTime): self.coords[:,2] += deltaTime; super().move(deltaTime) # ContourTrack
class ClickTrack(Track): # mouse down, then up                                   # ClickTrack
    def __init__(self, coords:np.ndarray, times:List[float]):                    # ClickTrack
        """Representing a mouse pressed and released event"""                    # ClickTrack
        super().__init__(*times); self.coords = coords # coords = [[x1, y1], [x2, y2]] # ClickTrack
    @staticmethod                                                                # ClickTrack
    def parse(events) -> List["ClickTrack"]:                                     # ClickTrack
        tracks = []; pressedEvents = defaultdict(lambda: None) # haha, get it?   # ClickTrack
        def process(e):                                                          # ClickTrack
            _type, x, y, button, t = e                                           # ClickTrack
            pe = pressedEvents[button]                                           # ClickTrack
            if _type == "mousePressed":                                          # ClickTrack
                if pe: raise Exception("Strange case. Why would inRange be true when mouse has just been pressed?") # ClickTrack
                pressedEvents[button] = e                                        # ClickTrack
            if _type == "mouseReleased":                                         # ClickTrack
                if pe: tracks.append(ClickTrack(np.array([pe[1:4], e[1:4]]), [pe[4], e[4]])); pressedEvents[button] = None # ClickTrack
                else: warnings.warn("Strange case. Why would mouse be released right at the start? Not strange enough to warrant an exception though") # ClickTrack
        events | filt(lambda x: x["type"] == "mousePressed" or x["type"] == "mouseReleased") | apply(lambda x: [x["type"], x["x"], x["y"], x["button"], x["timestamp"]/1000]) | apply(process) | deref() # ClickTrack
        return tracks                                                            # ClickTrack
    def isClick(self, threshold=1):                                              # ClickTrack
        """Whether this ClickTrack represents a single click.

:param threshold: if Manhattan distance between start and end is less than this amount, then declare it a single click""" # ClickTrack
        return abs(self.coords[0] - self.coords[1]).sum() <= threshold           # ClickTrack
    def __repr__(self): return f"<ClickTrack {self._displayTimes()} coords ({self.coords[0]} -> {self.coords[1]})>" # ClickTrack
    def _tooltip(self, ctx): return escapeHtml(f"{self}")                        # ClickTrack
    def events(self):                                                            # ClickTrack
        xy1, xy2 = self.coords; t1, t2 = self.timeUnix()                         # ClickTrack
        return [{"type": "mousePressed", "x": xy1[0], "y": xy1[1], "button": xy1[2], "timestamp": int(t1*1000)}, # ClickTrack
                {"type": "mouseReleased", "x": xy2[0], "y": xy2[1], "button": xy2[2], "timestamp": int(t2*1000)}] # ClickTrack
    def copy(self): return ClickTrack(self.coords | deref(), self.timeUnix())    # ClickTrack
class WheelTrack(Track):                                                         # WheelTrack
    def __init__(self, coords:np.ndarray, times:List[float]):                    # WheelTrack
        """Representing mouse wheel moved event"""                               # WheelTrack
        super().__init__(*times); self.coords = coords                           # WheelTrack
    @staticmethod                                                                # WheelTrack
    def parse(events) -> List["WheelTrack"]:                                     # WheelTrack
        d = []; _d = []; lastTime = 0                                            # WheelTrack
        for rot, t in events | filt(op()["type"] == "mouseWheelMoved") | apply(lambda x: [x["wheelRotation"], x["timestamp"]/1000]): # WheelTrack
            if t > lastTime + 2: d.append(_d); _d = []                           # WheelTrack
            _d.append([rot, t]); lastTime = t                                    # WheelTrack
        d.append(_d); return d | filt(lambda x: len(x)) | apply(aS(np.array) & (cut(1) | rows(0, -1)) | ~aS(WheelTrack)) | aS(list) # WheelTrack
    def __repr__(self): return f"<WheelTrack {self._displayTimes()} rotations (avg {self.coords[:,0].sum()}, {self.coords[:,0] | apply(lambda x: '+' if x > 0.5 else '0') | join('')})>" # WheelTrack
    def _tooltip(self, ctx): return escapeHtml(f"{self}")                        # WheelTrack
    def events(self):                                                            # WheelTrack
        rs = self.coords[:,0]; ts = np.linspace(*self.timeUnix(), self.coords.shape[0]) # WheelTrack
        return [rs, ts] | transpose() | ~apply(lambda rot, t: {"type": "mouseWheelMoved", "wheelRotation": rot, "timestamp": int(t*1000)}) # WheelTrack
    def copy(self): return WheelTrack(self.coords, self.timeUnix())              # WheelTrack
class StreamTrack(Track):                                                        # StreamTrack
    def __init__(self, frames:np.ndarray, times:np.ndarray):                     # StreamTrack
        """Representing screenshots from the UDP stream"""                       # StreamTrack
        super().__init__(times[0], times[-1]); self.frames = frames; self.times = times; self.aspect = self.frames.shape[2]/self.frames.shape[1] # StreamTrack
    @staticmethod                                                                # StreamTrack
    def parse(events) -> List["StreamTrack"]:                                    # StreamTrack
        events = events | filt(op()["type"] == "stream") | aS(list)              # StreamTrack
        if len(events) == 0: return []                                           # StreamTrack
        return [StreamTrack(*events | apply(lambda x: [x["frame"], x["timestamp"]/1000]) | transpose() | apply(np.array))] # StreamTrack
    def __repr__(self): return f"<StreamTrack {self._displayTimes()} #frames ({self.frames.shape[0]}) resolution {self.frames.shape[1:3][::-1]}>" # StreamTrack
    def _frames(self, n, f=iden()): return [self.frames, self.times] | transpose() | insertIdColumn(True, False) | f | aS(list) | aS(lambda x: x | batched(len(x)//n)) | item().all() # StreamTrack
    def _carousel(self): return self._frames(36) | cut(0) | toImg().all() | batched(9) | plotImgs(3, self.aspect, 3, im=True).all() | aS(k1.viz.Carousel) # StreamTrack
    def _tooltip(self, ctx):                                                     # StreamTrack
        metaId = ctx.metaId; streamId = autoId(); f = filt(ctx.dis1<op()<ctx.dis2, 1) # StreamTrack
        data = self._frames(40, f) | apply(toImg() | aS(k1.viz.HtmlImage, style="width:800px") | aS(lambda x: x._repr_html_()), 0) | deref() | aS(json.dumps) # StreamTrack
        ctx.scriptTags[streamId] = f"""
data_{streamId} = {data};
meta_{metaId}.cbs[{streamId}] = (x) => {{
    const stream_{streamId} = document.querySelector("#stream_{streamId}");
    const streamText_{streamId} = document.querySelector("#streamText_{streamId}");
    if (!stream_{streamId}) return;
    const fT = x/800*{ctx.dis2-ctx.dis1}+{ctx.dis1}; // frame time
    let minT = Infinity; let minIm = null; let minI = null
    for (const [imE, t, i] of data_{streamId}) {{
        const dT = Math.abs(fT-t);
        if (dT < minT) {{ minIm = imE; minT = dT; minI = i }}
        else break;
    }}
    stream_{streamId}.innerHTML = minIm;
    streamText_{streamId}.innerHTML = "frame: " + minI;
}};"""                                                                           # StreamTrack
        return f"""<div>{escapeHtml(str(self))}
            <div style="position:relative">
                <div id="stream_{streamId}"></div>
                <div id="streamText_{streamId}" style="position:absolute;top:8px;left:12px;padding:4px 8px;background-color:white;border-radius:12px"></div>
            </div>
        </div>"""                                                                # StreamTrack
    def _repr_html_(self): return f"""<div>{escapeHtml(str(self))}<div>{self._carousel()._repr_html_()}</div></div>""" # StreamTrack
    def events(self): return []                                                  # StreamTrack
    def copy(self): return StreamTrack(np.copy(self.frames), np.copy(self.times)) # StreamTrack
    def move(self, deltaTime): self.times += deltaTime; super().move(deltaTime)  # StreamTrack
def createTrackss(rec:Recording):                                                # createTrackss
    dis1 = rec.dis1; dis2 = rec.dis2; delta = dis2-dis1 # nTrack for "new track" # createTrackss
    def process(f=iden()):                                                       # createTrackss
        trackss = []                                                             # createTrackss
        for nTrack in rec._tracks | f | apply(lambda x: [max(x.startTime or 0, dis1+delta*0.01), min(x.endTime, dis2-delta*0.01), x]) | filt(op()>dis1, 1) | filt(op()<dis2, 0) | deref(): # createTrackss
            cTracks = None # "chosen track"                                      # createTrackss
            for eTracks in trackss: # "existing track"                           # createTrackss
                if eTracks["tracks"][-1][1] < nTrack[0]: cTracks = eTracks; break # can fit # createTrackss
            if cTracks: cTracks["tracks"].append(nTrack)                         # createTrackss
            else: trackss.append({"tracks": [nTrack], "type": nTrack[2].__class__.__name__.split(".")[-1]}) # createTrackss
        return trackss                                                           # createTrackss
    trackss = [                                                                  # createTrackss
        *process(instanceOf(CharTrack)),                                         # createTrackss
        *process(instanceOf(WordTrack)),                                         # createTrackss
        *process(instanceOf(ContourTrack)),                                      # createTrackss
        *process(instanceOf(ClickTrack)),                                        # createTrackss
        *process(instanceOf(WheelTrack)),                                        # createTrackss
        *process(instanceOf(StreamTrack))                                        # createTrackss
    ];                                                                           # createTrackss
    return [trackss, rec]                                                        # createTrackss
autoId = k1.AutoIncrement(random.randint(0, int(1e9)))                           # createTrackss
def drawTrackss(obj) -> "html":                                                  # drawTrackss
    h = settings.draw.trackHeight; pad = settings.draw.pad; trackss, rec = obj; sidebarW=120; # width # drawTrackss
    infoId = autoId(); metaId = autoId(); timeId = autoId(); timeLId = autoId(); sketchId = autoId(); sketchLId = autoId() # drawTrackss
    ctx = k1.Object.fromDict({"id2Tt": {}, "dis1": rec.dis1, "dis2": rec.dis2, "metaId": metaId, "scriptTags": {}}) # drawTrackss
                                                                                 # drawTrackss
    children = enumerate(trackss) | permute(1, 0) | ~apply(drawTracks, ctx=ctx) | join("") # drawTrackss
    trackNames = trackss | op()["type"].all() | insertIdColumn() | ~apply(lambda i, x: f"<div style='position:absolute;top:{pad+(pad+h)*i}px;left:12px;height:{h}px;text-align:center;line-height:{h}px'><div>{x}s</div></div>") | join("") # drawTrackss
    st0 = rec.dis1 - rec.startTime; et0 = rec.dis2 - rec.startTime; ticks0 = k1.ticks(st0, et0) # 0-based # drawTrackss
    ticksP = (ticks0+rec.startTime-rec.dis1)/(rec.dis2-rec.dis1)*800 # pixel scale # drawTrackss
    ticks = [ticks0, ticksP] | transpose() | filt(op()>0, 1) | filt(op()<800, 1) | ~apply(lambda x, y: f"<div style='position:absolute;width:1px;height:10px;background-color:black;left:{y}px;bottom:4px'></div>   <div style='position:absolute;left:{y-8}px;top:0px'>{x}</div>") | join("") # drawTrackss
    sketchH = (pad+h)*len(trackss)+pad; extraScripts = "\n".join(ctx.scriptTags.values()) # drawTrackss
    return f"""
<div style="display:flex;flex-direction:column;align-items:flex-start">
    <div style="display:flex;flex-direction:row">
        <div style="width:{sidebarW}px;padding-right:10px;display:flex;justify-content:center;align-items:center"><div>Time (s)</div></div>
        <div id="time_{timeId}" style="background-color:red;height:{h}px;position:relative;height:34px">
            {ticks}
            <div id="timeL_{timeLId}" style="position:absolute;top:0px;background-color:white;border:1px solid black;border-radius:8px;padding:0px 8px">&nbsp;&nbsp;</div>
        </div>
    </div>
    <div style="display:flex;flex-direction:row">
        <div style="width:{sidebarW}px;padding-right:10px;position:relative">{trackNames}</div>
        <div id="sketch_{sketchId}" style="width:{800}px;height:{sketchH}px;background-color:grey;position:relative">
            <div id="sketchL_{sketchLId}" style="position:absolute;width:1px;height:{sketchH}px;background-color:black;top:0px"></div>
            {children}
        </div>
    </div>
    <div id="info_{infoId}" style="min-height:30px;display:flex;flex-direction:column;justify-content:center;align-items:flex-start;padding:4px 12px"></div>
</div>
<script>
    id2Tt = {ctx.id2Tt | aS(json.dumps)}
    info_{infoId} = document.querySelector("#info_{infoId}");
    time_{timeId} = document.querySelector("#time_{timeId}");
    sketch_{sketchId} = document.querySelector("#sketch_{sketchId}");
    sketchL_{sketchLId} = document.querySelector("#sketchL_{sketchLId}");
    timeL_{timeLId} = document.querySelector("#timeL_{timeLId}");
    meta_{metaId} = {{x: 0, y: 0, cbs: {{}}}};
    for (const [k, v] of Object.entries(id2Tt)) {{
        let elem = document.querySelector(`#track_${{k}}`);
        elem.onmouseover = () => {{info_{infoId}.innerHTML = atob(v[0]);elem.style.backgroundColor = "red";}};
        elem.onmouseout  = () => {{info_{infoId}.innerHTML = "";        elem.style.backgroundColor = "white";}};
    }}
    sketch_{sketchId}.onmousemove = (event) => {{
        const x = event.pageX-sketch_{sketchId}.getBoundingClientRect().x;
        meta_{metaId}.x = x;
        sketchL_{sketchLId}.style.left = x + "px";
        timeL_{timeLId}.style.left = (x-timeL_{timeLId}.getBoundingClientRect().width/2) + "px";
        timeL_{timeLId}.innerHTML = Number(x/800*{et0-st0}+{st0}).toFixed(2) + "s";
        for (const cb of Object.values(meta_{metaId}.cbs)) cb(x);
    }}
    {extraScripts}
</script>"""                                                                     # drawTrackss
def drawTracks(tracks, rowId, ctx) -> "html": return tracks["tracks"] | apply(drawTrack, rowId=rowId, ctx=ctx) | join("") # drawTracks
def drawTrack(track, rowId, ctx) -> "html":                                      # drawTrack
    h = settings.draw.trackHeight; pad = settings.draw.pad; st, et, obj = track  # drawTrack
    x1 = (st-ctx.dis1)/(ctx.dis2-ctx.dis1)*800; x2 = (et-ctx.dis1)/(ctx.dis2-ctx.dis1)*800 # drawTrack
    y = rowId*(h+pad)+pad; w = x2-x1; trackId = autoId()                         # drawTrack
    tooltip = obj._tooltip(ctx).encode() | aS(base64.b64encode) | op().decode()  # drawTrack
    ctx.id2Tt[trackId] = [tooltip, x1, x2, y]                                    # drawTrack
    return f"""<div id="track_{trackId}" style="top:{y}px;left:{x1}px;width:{w}px;height:{h}px;background-color:white;position:absolute"></div>""" # drawTrack
basePath = os.path.dirname(inspect.getabsfile(k1lib)) + os.sep + "k1ui" + os.sep # drawTrack
@k1.patch(Recording, static=True)                                                # drawTrack
def sampleEvents() -> List[dict]:                                                # sampleEvents
    """Grabs the built-in example events. Results will be really long,
so beware, as it can crash your notebook if you try to display it."""            # sampleEvents
    mouseE, keyE = cat(f"{basePath}mouseKey.pth", False) | aS(dill.loads)        # sampleEvents
    deltaT = keyE()[0]["timestamp"] - mouseE()[0]["timestamp"]                   # sampleEvents
    ev = [*mouseE() | apply(lambda x: {**x, "timestamp": x["timestamp"]+deltaT}), *keyE()] # sampleEvents
    try: # local comp has the k1ui-screen file, but it will not be bundled with the library, cause it's like 80MB! # sampleEvents
        screenE = cat("screen.pth", False) | aS(dill.loads)                      # sampleEvents
        deltaT = keyE()[0]["timestamp"] - screenE()[0]["timestamp"]              # sampleEvents
        return [*screenE() | apply(lambda x: {**x, "timestamp": x["timestamp"]+deltaT}), *ev] # sampleEvents
    except: return ev                                                            # sampleEvents
@k1.patch(Recording, static=True)                                                # sampleEvents
def sample() -> Recording:                                                       # sample
    """Creates a Recording from :meth:`sampleEvents`"""                          # sample
    return Recording(Recording.sampleEvents())                                   # sample
@k1.patch(ContourTrack)                                                          # sample
def split(self, times:List[float]):                                              # split
    """Splits this contour track by multiple timestamps relative
to recording's start time. Example::

    r = k1ui.Recording.sample()
    r.sel1(klass=k1ui.ContourTrack).split([5])"""                                # split
    rec = self.recording; c = self.coords; i = 0; x = 0; y = 0; d = []; cps = np.array(times) + rec.startTime # split
    while True:                                                                  # split
        if cps[i] > c[y,2]: y += 1                                               # split
        else:                                                                    # split
            if y > x: d.append(c[x:y])                                           # split
            x = y; i += 1                                                        # split
        if y >= len(c): d.append(c[x:y]); break                                  # split
        if i >= len(cps): d.append(c[x:]); break                                 # split
    rec.removeTracks(self)                                                       # split
    rec.addTracks(d | apply(ContourTrack))                                       # split
@k1.patch(ContourTrack)                                                          # split
def splitClick(self, clickTracks:List["ClickTrack"]=None):                       # splitClick
    """Splits this contour track by click events. Essentially, the click
events chops this contour into multiple segments. Example::

    r = k1ui.Recording.sample()
    r.sel1(klass=k1ui.ContourTrack).splitClick()

:param clickTracks: if not specified, use all ClickTracks from the recording"""  # splitClick
    rec = self.recording; c = self.coords; i = 0; x = 0; y = 0; d = []           # splitClick
    if clickTracks is None: clickTracks = rec.sel(*self.time0Rec()) | instanceOf(ClickTrack) # splitClick
    self.split(clickTracks | ~filt(op().isClick(-1)) | op().timeUnix().all() | joinStreams() | sort(None) | apply(op()-rec.startTime) | deref()) # splitClick
@k1.patch(Recording)                                                             # splitClick
def addTime(self, t:float, duration:float) -> Recording:                         # addTime
    """Inserts a specific duration into a specific point in time.
More clearly, this transfroms this::

    # |-1--|   |-2-|
    #    |---3---|
    #         ^ insert duration=3 here

Into this::

    # |-1--|      |-2-|
    #    |---3------|

Tracks that partly overlaps with the range will have their start/end times
modified, and potentially delete some of the Track's internal data:

- Tracks whose only start and end times are modified: Char, Word, Click, Wheel
- Tracks whose internal data are also modified: Contour, Stream

:param t: where to insert the duration, relative to Recording's start time
:param duration: how long (in seconds) to insert?"""                             # addTime
    at = self.sel(t,t); after = self.sel(t) # tracks at or after the specified time # addTime
    unix = t + self.startTime                                                    # addTime
    for track in at: after.remove(track)                                         # addTime
    for track in at:                                                             # addTime
        track.endTime += duration                                                # addTime
        if isinstance(track, ContourTrack):                                      # addTime
            c = track.coords; idx = (c[:,2] > unix).argmax(); track._cachedImg = None # addTime
            if c[idx,2] > unix: c[idx:,2] += duration # index is valid           # addTime
        if isinstance(track, StreamTrack):                                       # addTime
            c = track.times; idx = (c > unix).argmax()                           # addTime
            if c[idx]   > unix: c[idx:]   += duration # index is valid           # addTime
    for track in after:                                                          # addTime
        track.startTime += duration; track.endTime += duration                   # addTime
        if isinstance(track, ContourTrack): track.coords[2] += duration          # addTime
        if isinstance(track, StreamTrack): track.times += duration               # addTime
    self.endTime += duration; self._resetDis(); return self                      # addTime
@k1.patch(Recording)                                                             # addTime
def removeTime(self, t1:float, t2:float) -> Recording:                           # removeTime
    """Deletes time from t1 to t2 (relative to Recording's start time).
All tracks lying completely inside this range will be deleted. More
clearly, it transforms this::

    # |-1--|  |-2-|   |-3-|
    #    |---4---|  |-5-|
    #        ^       ^ delete between these carets

Into this::

    # |-1--|   |-3-|
    #    |-4-||5-|

Tracks that partly overlaps with the range will have their start/end times
modified, and potentially delete some of the Track's internal data:

- Tracks whose only start and end times are modified: Char, Word, Click, Wheel
- Tracks whose internal data are also modified: Contour, Stream"""               # removeTime
    duration = t2 - t1; t1U = t1 + self.startTime; t2U = t2 + self.startTime     # removeTime
    self.removeTracks(self.sel(t1, t2) | filt(op().startTime >= t1U) | filt(op().endTime < t2U)) # removing everything that's completely inside # removeTime
    overlap = self.sel(t1, t2) | aS(list); after = self.sel(t2) | filt(op().startTime >= t2U) | aS(list) # removeTime
    for track in overlap: # handling left overhang                               # removeTime
        if isinstance(track, ContourTrack):                                      # removeTime
            c = track.coords; idx1 = (c[:,2] > t1U).argmax(); idx2 = (c[:,2] > t2U).argmax() # removeTime
            if c[idx2,2] <= t2U: idx2 = len(c)                                   # removeTime
            a = c[:idx1]; b = c[idx2:]; b[:,2] -= duration                       # removeTime
            track.coords = np.concatenate([a, b]); track._cachedImg = None       # removeTime
        if isinstance(track, StreamTrack):                                       # removeTime
            c = track.times; idx1 = (c > t1U).argmax(); idx2 = (c > t2U).argmax() # removeTime
            if c[idx2] <= t2U: idx2 = len(c) # special case if idx2 is not valid # removeTime
            track.times = np.concatenate([track.times[:idx1], track.times[idx2:]-duration]) # removeTime
            track.frames = np.concatenate([track.frames[:idx1], track.frames[idx2:]]) # removeTime
        track.endTime = max(t1U, track.endTime - duration); track.startTime = min(t1U, track.startTime) # removeTime
    for track in after:                                                          # removeTime
        if isinstance(track, ContourTrack): track.coords[:,2] -= duration        # removeTime
        if isinstance(track, StreamTrack): track.times -= duration               # removeTime
        track.startTime -= duration; track.endTime -= duration                   # removeTime
    self._resetTimes(); self._resetDis(); return self                            # removeTime
def _move(cs, e1, e2):                                                           # _move
    det = e1[0]*e2[1] - e1[1]*e2[0]; dot = e1@e2; angle = math.atan2(det, dot)   # _move
    s = math.sin(angle); c = math.cos(angle); rot = np.array([[c, -s], [s, c]])  # _move
    scale = (e2**2).sum()**0.5/(e1**2).sum()**0.5; return (rot @ cs.T)*scale | transpose() # _move
@k1.patch(ContourTrack)                                                          # _move
def movePoint(self, x, y, start=True):                                           # movePoint
    """Move contour's start/end to another location, smoothly scaling all
intermediary points along.

:param start: if True, move the start point, else move the end point"""          # movePoint
    c = self.coords; e2 = np.array([x, y])                                       # movePoint
    if start: s = c[-1,:2]; e1 = c[0,:2] - s                                     # movePoint
    else: s = c[0,:2]; e1 = c[-1,:2] - s                                         # movePoint
    e2 = e2 - s; c[:,:2] = _move(c[:,:2]-s, e1, e2)+s                            # movePoint
@k1.patch(Track)                                                                 # movePoint
def nextTrack(self) -> Track:                                                    # nextTrack
    """Grabs the next track (ordered by start time) in the recording"""          # nextTrack
    return self.recording._tracks | filt(op().startTime) | filt(op().startTime > (self.startTime or 0)) | sortF(op().startTime) | item() # nextTrack
@k1.patch(Recording)                                                             # nextTrack
def refine(self, enabled:List[int]=[1,1,0]) -> Recording:                        # refine
    """Perform sensible default operations to refine the Recording.
This currently includes:

- (0) Splitting ContourTracks into multiple smaller tracks using click events
- (1) Forming words from nearby CharTracks
- (2) Removing open-close CharTracks. Basically, CharTracks that don't have a begin or end time

:param enabled: list of integers, whether to turn on or off certain features. 1 to turn on, 0 to turn off""" # refine
    if enabled[0]: self.formWords()                                              # refine
    if enabled[1]: self.sel(klass=ContourTrack) | op().splitClick().all() | ignore() # refine
    if enabled[2]: self.removeTracks(self.sel(klass=CharTrack) | ~filt(op().startTime)) # refine
    return self                                                                  # refine
def convBlock(inC, outC, kernel=3, stride=2, padding=1):                         # convBlock
    return torch.nn.Sequential(torch.nn.Conv2d(inC, outC, kernel, stride, padding), torch.nn.ReLU(), torch.nn.BatchNorm2d(outC)) # convBlock
if hasTorch:                                                                     # convBlock
    class skipBlock(torch.nn.Module):                                            # convBlock
        def __init__(self, inC):                                                 # convBlock
            super().__init__(); self.conv1 = convBlock(inC, inC, stride=1)       # convBlock
            self.conv2 = convBlock(inC, inC*2)                                   # convBlock
        def forward(self, x): return ((x | self.conv1) + x) | self.conv2         # convBlock
    class Net(torch.nn.Module):                                                  # convBlock
        def __init__(self, skips:int=5):                                         # convBlock
            super().__init__()                                                   # convBlock
            self.skips = torch.nn.Sequential(convBlock(3, 8), *[skipBlock(8*2**i) for i in range(skips)]) # convBlock
            self.avgPool = torch.nn.AdaptiveAvgPool2d([1, 1]); self.lin1 = knn.LinBlock(8 * 2**skips, 50) # convBlock
            self.lin2 = torch.nn.Linear(50, 10); self.softmax = torch.nn.Softmax(dim=1) # convBlock
            self.distThreshold = torch.nn.Parameter(torch.tensor(-0.5)); self.sigmoid = torch.nn.Sigmoid() # convBlock
            self.headOnly = True                                                 # convBlock
        def forward(self, x):                                                    # convBlock
            x = x | self.skips | self.avgPool | op().squeeze() | self.lin1       # convBlock
            return x if self.headOnly else x | self.lin2                         # convBlock
            x = ((x[None] - x[:,None])**2).sum(dim=-1)                           # convBlock
            x = (x + 1e-7)**0.5 + self.distThreshold | self.sigmoid              # convBlock
            return x                                                             # convBlock
def distNet() -> "torch.nn.Module":                                              # distNet
    """Grabs a pretrained network that might be useful in distinguishing
between screens. Example::

    net = k1ui.distNet()
    net(torch.randn(16, 3, 192, 192)) # returns tensor of shape (16, 10)"""      # distNet
    net = Net(); net.load_state_dict(cat(f"{basePath}256.model.state_dict.pth", False) | aS(dill.loads)) # distNet
    net.parameters() | op().requires_grad_(False).all() | ignore(); net.eval(); return net # distNet
def discardTransients(it, col=None, countThres=7, regular=False): # consistent for 7 consecutive frames, then output the results # discardTransients
    lastRow = None; lastE = None                                                 # discardTransients
    yielded = False; count = 0                                                   # discardTransients
    for row in it:                                                               # discardTransients
        e = row[col] if col else row                                             # discardTransients
        if e == lastE: count += 1                                                # discardTransients
        else: count = 0; lastE = e; lastRow = row; yielded = False               # discardTransients
        if count > countThres-2 and not yielded: yielded = True; yield lastRow   # discardTransients
        elif regular: yield None                                                 # discardTransients
class Buffer:                                                                    # Buffer
    def __init__(self): self.l = deque()                                         # Buffer
    def append(self, x): self.l.append(x)                                        # Buffer
    def __next__(self): return self.l.popleft()                                  # Buffer
if hasTorch and hasTv:                                                           # Buffer
    np2Tensor = toImg() | aS(tf.Resize([192, 192])) | toTensor()                 # Buffer
    class MLP(nn.Module):                                                        # Buffer
        def __init__(self, nClasses, **kwargs):                                  # Buffer
            super().__init__(); self.l1 = knn.LinBlock(50, nClasses); self.l2 = nn.Linear(nClasses, nClasses) # Buffer
        def forward(self, xb): return xb | self.l1 | self.l2                     # Buffer
whatever = object()                                                              # Buffer
class TrainScreen:                                                               # TrainScreen
    data: List[Tuple[int, str]]                                                  # TrainScreen
    """Core dataset of TrainScreen. Essentially just a list of (frameId, screen name)""" # TrainScreen
    def __init__(self, r:Recording):                                             # TrainScreen
        """Creates a screen training system that will train a small neural network
to recognize different screens using a small amount of feedback from the user.
Overview on how it's supposed to look like:

Setting up::

    r = k1ui.Recording(await k1ui.record(30)) # record everything for 30 seconds, and creates a recording out of it
    ts = k1ui.TrainScreen(r) # creates the TrainScreen object
    r # run this in a cell to display the recording, including StreamTrack
    ts.addRule("home", "settings", "home") # add expected screen transition dynamics (home -> settings -> home)

Training with user's feedback::

    ts.registerFrames({"home": [100, 590, 4000, 4503], "settings": [1200, 2438]}) # label some frames of the recording. Network will train for ~6 seconds
    next(ts) # display 20 images that confuses the network the most
    ts.register({"home": [2, 6], "settings": [1, 16]}) # label some frames from the last line. Notice the frame numbers are much smaller and are <20
    next(ts); ts.register({}); next(ts); ts.register({}) # repeat the last 2 lines for a few times (3-5 times is probably good enough for ~7 screens)

Evaluating the performance::

    ts.graphs() # displays 2 graphs: network's prediction graph and the actual rule graph. Best way to judge performance
    ts.l.Accuracy.plot() # actual accuracy metric while training. Network could have bad accuracy here while still able to construct a perfect graph, so don't rely much on this

Using the model::

    ts.predict(torch.randn(2, 3, 192, 192) | k1ui.distNet()) # returns list of ints. Can use ts.idx2Name dict to convert to screen names

Saving the model::

    ts | aS(dill.dumps) | file("ts.pth")

.. warning::

    This won't actually save the associated recording, because recordings are
    very heavy objects (several GB). It is expected that you manually manage
    the lifecycle of the recording."""                                           # TrainScreen
        self.r = r; self.data = []; # [(frame id, screen name)]                  # TrainScreen
        self._aspect = self.frames | item() | op().shape[:2] | ~aS(lambda x, y: y/x) # TrainScreen
        self._distNet = distNet(); self._rules = set(); self._trainParams = {"joinAlpha": 0, "epochs": 300} # TrainScreen
        self._lastScreenName = None; self._screenDump = Buffer(); self._screenTransients = discardTransients(self._screenDump, regular=True) # TrainScreen
    @property                                                                    # TrainScreen
    def _coldStart(self): return len(self.data) == 0 # whether there are any data at all to work with # TrainScreen
    def _coldGuard(self):                                                        # TrainScreen
        if self._coldStart: raise Exception("TrainScreen has not started yet. Run `next(ts)`, choose a few frames using `ts.register()` to access this functionality") # TrainScreen
    def _learner(self):                                                          # TrainScreen
        self._coldGuard(); l = k1.Learner(); l.data = self._dataF()              # TrainScreen
        l.model = MLP(len(self.name2Idx))                                        # TrainScreen
        l.opt = optim.AdamW([l.model.parameters(), self._distNet.parameters()] | joinStreams(), lr=3e-3) # TrainScreen
        l.cbs.add(Cbs.LossCrossEntropy()); l.css = "none"                        # TrainScreen
        l.ConfusionMatrix.categories = deref()(self.name2Idx.items()) | sort(1) | cut(0) | deref() # TrainScreen
        l.cbs.remove("AccuracyTop5", "AccF0"); return l                          # TrainScreen
    def train(self, restart=True):                                               # TrainScreen
        """Trains the network for a while (300 epochs/6 seconds). Will be called
automatically when you register new frames to the system

:param restart: whether to restart the small network or not"""                   # TrainScreen
        if restart: self.l = self._learner();                                    # TrainScreen
        self.l.run(self._trainParams["epochs"])                                  # TrainScreen
    def trainParams(self, joinAlpha:float=None, epochs:int=None):                # TrainScreen
        """Sets training parameters.

:param joinAlpha: (default 0) alpha used in joinStreamsRandom component for each
    screen categories. Read more at :class:`~k1lib.cli.structural.joinStreamsRandom`
:param epochs: (default 300) number of epochs for each training session"""       # TrainScreen
        if joinAlpha: self._trainParams["joinAlpha"] = joinAlpha                 # TrainScreen
        if epochs: self._trainParams["epochs"] = epochs                          # TrainScreen
    @property                                                                    # TrainScreen
    def frames(self) -> np.ndarray:                                              # TrainScreen
        """Grab the frames from the first :class:`StreamTrack` from the :class:`Recording`""" # TrainScreen
        return self.r.sel1(klass=StreamTrack).frames                             # TrainScreen
    @property                                                                    # TrainScreen
    @lru_cache                                                                   # TrainScreen
    def feats(self) -> List[np.ndarray]:                                         # TrainScreen
        """Gets the feature array of shape (N, 10) by passing the frames
through :meth:`distNet`. This returns a list of arrays, not a giant,
stacked array for memory performance"""                                          # TrainScreen
        self._coldGuard(); print("Converting all frames to features using `distNet`..."); a = k1.AutoIncrement() # TrainScreen
        res = self.frames | tee(lambda x: f"{a()}/{len(self)}").crt() | np2Tensor.all() | batched(16, True) | apply(aS(list) | aS(torch.stack) | aS(self._distNet)) | joinStreams() | aS(list) # TrainScreen
        print(); return res                                                      # TrainScreen
    def __len__(self): return len(self.frames)                                   # TrainScreen
    def _randomConsidering(self): return range(len(self)) | splitW(1, 1, 1, 1, 1) | apply(randomize(None, 42) | head(4)) | joinStreams() | aS(list) # TrainScreen
    def __next__(self) -> "PIL.Image.Image": # show frames                       # TrainScreen
        if self._coldStart: self._considering = self._randomConsidering()        # TrainScreen
        else:                                                                    # TrainScreen
            a = self.transitionScreens(False) | randomize(None, 42) | cut(0) | aS(iter) # TrainScreen
            b = self._randomConsidering() | aS(iter); c = self._midBoundaryConsidering() | aS(iter) # TrainScreen
            self._considering = [a, a, b, c, c, c] | apply(wrapList() | insert(yieldT | repeat(), False) | joinStreams() | randomize()) | joinStreamsRandom() | head(20) | deref() # TrainScreen
        return self._considering | lookup(self.frames) | insertIdColumn(begin=False) | plotImgs(5, self._aspect-0.2, im=True) # TrainScreen
    def _refreshIdx(self): self.idx2Name, self.name2Idx = self.data | cut(1) | aS(set) | insertIdColumn() | toDict() & (permute(1, 0) | toDict()) | deref() # TrainScreen
    def register(self, d):                                                       # TrainScreen
        """Tells the object which images previously displayed by :meth:`__next__`
associate with what screen name. Example::

    next(ts) # displays the images out to a notebook cell
    ts.register({"home": [3, 4, 7], "settings": [5, 19, 2], "monkeys": [15, 11], "guns": []})

This will also quickly (around 6 seconds) train a small neural network on all
available frames based on the new information you provided.

See also: :meth:`registerFrames`"""                                              # TrainScreen
        self.data = [self.data, deref()(d.items()) | apply(repeat(), 0) | transpose().all() | joinStreams() | permute(1, 0) | lookup(self._considering, 0)] | joinStreams() | sort(0) | unique(0) | deref() # TrainScreen
        self._refreshIdx(); self.train()                                         # TrainScreen
    def registerFrames(self, data:Dict[str, List[int]]):                         # TrainScreen
        """Tells the object which frames should have which labels.
Example::

    ts.registerFrames({"home": [328, 609], "settings": [12029], "monkeys": [1238]})

This differs from :meth:`register` in that the frame id here is the
absolute frame index in the recording, while in :meth:`register`,
it's the frame displayed by :meth:`__next__`."""                                 # TrainScreen
        self.data = [self.data, deref()(data.items()) | apply(repeat(), 0) | transpose().all() | joinStreams() | permute(1, 0)] | joinStreams() | sort(0) | unique(0) | deref(); self._refreshIdx(); self.train() # TrainScreen
    def addRule(self, *screenNames:List[str]) -> "TrainScreen":                  # TrainScreen
        """Adds a screen transition rule. Let's say that the transition
dynamic looks like this:

.. code-block:: text

    home <---> settings <---> account
                  ^
                  |
                  v
              shortcuts

You can represent it like this::

    ts.addRule("home", "settings", "account", "settings", "home")
    ts.addRule("settings", "shortcuts", "settings")"""                           # TrainScreen
        screenNames | window(2) | apply(tuple) | apply(self._rules.add) | ignore(); return self # TrainScreen
    def transitionScreens(self, obeyRule:bool=whatever) -> List[Tuple[int, str]]: # TrainScreen
        """Get the list of screens (list of (frameId, screen name) tuple) that
the network deems to be transitions between screen states.

:param obeyRule: if not specified, then don't filter. If True, returns only screens that
    are part of the specified rule and vice versa"""                             # TrainScreen
        self._coldGuard()                                                        # TrainScreen
        with torch.no_grad(): transitions = self.predict(self.feats) | insertIdColumn() | aS(discardTransients, 1) | window(2, True, None) | filt(~aS(lambda x, y: not y or x[1] != y[1])) | cut(0) | lookup(self.idx2Name, 1) | deref() # TrainScreen
        if obeyRule is whatever: return transitions                              # TrainScreen
        f = inSet(self._rules, 1) if obeyRule else ~inSet(self._rules, 1)        # TrainScreen
        return transitions | window(2) | apply(transpose() | iden() + aS(tuple)) | f | transpose().all() | joinStreams() | unique(0) | deref() # TrainScreen
    def newEvent(self, sess:WsSession, event:dict):                              # TrainScreen
        if event["type"] == "stream":                                            # TrainScreen
            with torch.no_grad():                                                # TrainScreen
                name = event["frame"] | np2Tensor | op().reshape(-1, 3, 192, 192) | aS(self._distNet) | op().view(1, -1)\
                    | self.l.model | op().argmax().item() | aS(lambda x: self.idx2Name[x]) # TrainScreen
                sess.loop.create_task(sess.eventCb(sess, {"type": "screenName", "name": name})) # TrainScreen
        if event["type"] == "screenName":                                        # TrainScreen
            self._screenDump.append(event["name"]); res = next(self._screenTransients) # TrainScreen
            if res:                                                              # TrainScreen
                sess.loop.create_task(sess.eventCb(sess, {"type": "screenTransition", "transition": (self._lastScreenName, res)})) # TrainScreen
                self._lastScreenName = res                                       # TrainScreen
    def predict(self, feats:"torch.Tensor") -> List[int]:                        # TrainScreen
        """Using the built-in network, tries to predict the screen name for a
bunch of features of shape (N, 10). Example::

    r = ...; ts = k1ui.TrainScreen(r); next(ts)
    ts.register({"bg": [9, 10, 11, 12, 17, 19], "docs": [5, 6, 7, 8, 0, 1, 4], "jupyter": [2, 3]})
    # returns list of 2 integers
    ts.predict(torch.randn(2, 3, 192, 192) | aS(k1ui.distNet()))"""              # TrainScreen
        self._coldGuard(); return feats | batched(128, True) | apply(aS(list) | aS(torch.stack) | aS(self.l.model) | op().argmax(1).numpy()) | joinStreams() # TrainScreen
    def transitionGraph(self) -> "graphviz.dot.Digraph":                         # TrainScreen
        """Gets a screen transition graph of the entire recording. See also: :meth:`graphs`""" # TrainScreen
        g = k1.digraph(); self.transitionScreens() | cut(1) | window(2) | apply(tuple) | count() | cut(0, 1) | ~apply(lambda c, xy: g(*xy, label=f" {c}")) | ignore(); return g # TrainScreen
    def ruleGraph(self) -> "graphviz.dot.Digraph":                               # TrainScreen
        """Gets a screen transition graph based on the specified rules.
Rules are added using :meth:`addRule`. See also: :meth:`graphs`"""               # TrainScreen
        g = k1.digraph(); self._rules | ~apply(g) | ignore(); return g           # TrainScreen
    def graphs(self) -> viz.Carousel:                                            # TrainScreen
        """Combines both graphs from :meth:`transitionGraph` and :meth:`ruleGraph`""" # TrainScreen
        return [self.transitionGraph(), self.ruleGraph()] | toImg().all() | aS(viz.Carousel) # TrainScreen
    def labeledData(self) -> viz.Carousel:                                       # TrainScreen
        """Visualizes labeled data"""                                            # TrainScreen
        return self.data | groupBy(1) | apply(randomize(None) | head(5) | lookup(self.frames, 0)) | batched(5) | plotImgs(5, self._aspect-0.2, table=True, im=True).all() | aS(viz.Carousel) # TrainScreen
    def __getstate__(self): d = dict(self.__dict__); del d["r"]; del d["_lastScreenName"]; del d["_screenDump"]; del d["_screenTransients"]; return d # TrainScreen
    def __setstate__(self, d):                                                   # TrainScreen
        self.__dict__.update(d); self._lastScreenName = None; self._screenDump = Buffer() # TrainScreen
        self._screenTransients = discardTransients(self._screenDump, regular=True) # TrainScreen
    def correctRatio(self):                                                      # TrainScreen
        """Ratio between the number of screens that is in a valid transition and
ones that isn't in a valid transition. Just a quick metric to see how well the
network is doing. The higher the number, the better it is"""                     # TrainScreen
        return len(self.transitionScreens(True))/len(self.transitionScreens(False)) # TrainScreen
def fillIn(n, states):                                                           # fillIn
    iS = 0 # index of states                                                     # fillIn
    state = None; nextI, nextS = states[iS]                                      # fillIn
    for i in range(n):                                                           # fillIn
        if i >= nextI:                                                           # fillIn
            iS += 1; state = nextS                                               # fillIn
            if iS < len(states): nextI, nextS = states[iS]                       # fillIn
        yield [i, state]                                                         # fillIn
def blocks(it):                                                                  # blocks
    lastIs = []; lastE = None                                                    # blocks
    for i, e in it:                                                              # blocks
        if e != lastE:                                                           # blocks
            if lastE is not None: yield min(lastIs), max(lastIs), lastE          # blocks
            lastIs = []; lastE = e                                               # blocks
        lastIs.append(i)                                                         # blocks
    yield min(lastIs), max(lastIs), lastE                                        # blocks
@k1.patch(TrainScreen)                                                           # blocks
def _dataF(self, bs=64):                                                         # _dataF
    self._coldGuard(); v1 = fillIn(len(self), self.data) | filt(op(), 1) # old version. A bit more liveral than v2, and will accidentally auto label wrongly from time to time # _dataF
    v2 = blocks(self.data) | ~apply(lambda x, y, z: [range(x, y+1), z | repeat()] | transpose()) | joinStreams() # _dataF
    js = deref() | aS(lambda xs: xs | apply(repeatFrom() | randomize()) | joinStreamsRandom(self._trainParams["joinAlpha"], xs | apply(len) | deref())) # joinStreams # _dataF
    return v2 | randomize(None) | groupBy(1) | filt(lambda x: len(x) > 1) | splitW().all() | transpose()\
        | apply(js | lookup(self.feats, 0) | lookup(self.name2Idx, 1) | batched(bs)\
                | apply(transpose() | (aS(list) | aS(torch.stack)) + toTensor(int)) # _dataF
               ) | stagger.tv(1024/bs) | aS(list)                                # _dataF
def midBounds(it): # to grab data samples that's in between blocks. Aka the really confusing case in-between transitions, so that the user can guide it effectively # midBounds
    lastI = 0; lastE = None                                                      # midBounds
    for i, e in it:                                                              # midBounds
        if e != lastE: yield (i + lastI)//2, i-lastI, lastE; lastE = e           # midBounds
        lastI = i                                                                # midBounds
    return it                                                                    # midBounds
@k1.patch(TrainScreen)                                                           # midBounds
def _midBoundaryConsidering(self): return midBounds(self.data) | ~head(1) | ~sort(1) | cut(0) # _midBoundaryConsidering
